"""Home assistant related plugins."""

import asyncio
import json

import asyncws

from pnp.plugins.pull import AsyncPull
from pnp.utils import make_list, include_or_exclude, wildcards_to_regex


class State(AsyncPull):
    """
    Connects to the home assistant websocket api and listens for state changes.
    If no include or exclude is defined it will report all state changes.
    If include is defined only entities that match one of the specified patterns will be emitted.
    If exclude if defined entities that match at least one of the specified patterns will
    be ignored. Exclude patterns overrides include patterns.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/hass.State/index.md
    """
    __REPR_FIELDS__ = ['exclude', 'include', 'url']

    def __init__(self, url, token, include=None, exclude=None, **kwargs):
        super().__init__(**kwargs)
        self.url = self._sanitize_url(url)
        self.token = str(token)
        self.include = make_list(include)
        self.exclude = make_list(exclude)
        self._websocket = None
        self._loop = None

        self._include_regex = self.include and wildcards_to_regex(self.include)
        self._exclude_regex = self.exclude and wildcards_to_regex(self.exclude)

    @staticmethod
    def _sanitize_url(url):
        res = str(url).replace('https://', 'wss://').replace('http://', 'ws://')
        if res.endswith('/'):
            return res[:-1]  # Remove last /
        return res

    @staticmethod
    def _layout_message(message):
        def _layout_state(state):
            from copy import copy
            _copy = copy(state)
            _copy.pop('entity_id', None)
            _copy.pop('context', None)
            return _copy

        state_data = message.get('event', {}).get('data', {})
        entity_id = state_data.get('entity_id')
        old_state = _layout_state(state_data.get('old_state', {}))
        new_state = _layout_state(state_data.get('new_state', {}))
        return {'entity_id': entity_id, 'old_state': old_state, 'new_state': new_state}

    async def _emit(self, message):
        payload = self._layout_message(message)
        if include_or_exclude(payload['entity_id'], self._include_regex, self._exclude_regex):
            self.notify(payload)

    async def _receive_states(self):
        self._websocket = await asyncws.connect('{self.url}/api/websocket'.format(**locals()))

        while True:
            message = await self._websocket.recv()
            if message is None:
                await self._websocket.close()
                self._websocket = None
                break
            message = json.loads(message)
            if message.get('type', '') == 'auth_required':
                hass_version = message.get('ha_version')
                self.logger.info(
                    "Connected to Home Assistant %s websocket @ %s",
                    hass_version, self.url
                )
                self.logger.debug("Authentication is required. Providing token")
                await self._websocket.send(json.dumps({
                    'type': 'auth',
                    'access_token': self.token
                }))
            elif message.get('type', '') == 'auth_ok':
                self.logger.info("Authentication is valid")
                await self._websocket.send(json.dumps(
                    {'id': 1, 'type': 'subscribe_events', 'event_type': 'state_changed'}
                ))
            elif message.get('type', '') == 'auth_invalid':
                self.logger.info("Authentication is invalid. Aborting...")
                await self._websocket.close()
            elif message.get('type', '') == 'result':
                pass
            elif message.get('type', '') == 'event':
                await self._emit(message)
            else:
                self.logger.warning("Got unexpected message '%s'", message)

    async def _stop(self):
        await super()._stop()
        if self._websocket:
            await self._websocket.close()

    async def _pull(self):
        self._loop = asyncio.get_event_loop()
        await self._receive_states()



"""MQTT related plugins."""

from pnp.plugins.pull import SyncPull


class Subscribe(SyncPull):
    """
    Pulls messages from the specified topic from the given mosquitto mqtt
    broker (identified by host and port).


    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/mqtt.Subscribe/index.md
    """
    __REPR_FIELDS__ = ['host', 'port', 'topic', 'user']

    def __init__(self, host, topic, port=1883, user=None, password=None, **kwargs):
        super().__init__(**kwargs)
        self.host = str(host)
        self.topic = str(topic)
        self.port = int(port)
        self.user = user and str(user)
        self.password = password and str(password)
        self._client = None

    def _on_connect(self, client, userdata, flags, rc):  # pylint: disable=unused-argument
        # Subscribing in on_connect() means that if we lose the connection and
        # reconnect then subscriptions will be renewed.
        if rc == 0:
            client.subscribe(self.topic)
            self.logger.info("Connected with result code '%s' "
                             "to %s @ %s:%s", rc, self.topic, self.host, self.port)
        else:
            self.logger.error("Bad connection with result code '%s' "
                              "to %s @ %s:%s", rc, self.topic, self.host, self.port)

    def _on_disconnect(self, client, userdata, rc):  # pylint: disable=unused-argument
        if rc != 0:
            self.logger.warning("Unexpected mqtt disconnect with result code '%s'. "
                                "Will automatically reconnect.", rc)

    def _on_message(self, client, obj, msg):  # pylint: disable=unused-argument
        self.logger.debug("Got message from broker on topic '%s'. "
                          "Payload='%s'", self.topic, msg.payload)

        # This one is an envelope with data
        self.notify(dict(
            topic=str(msg.topic),
            levels=msg.topic.split('/'),
            payload=msg.payload.decode('utf-8')
        ))

    def _stop(self):
        super()._stop()
        if self._client:
            self._client.disconnect()

    def _pull(self):
        import paho.mqtt.client as paho

        self._client = paho.Client()
        if self.user:
            self._client.username_pw_set(self.user, self.password)
        self._client.on_connect = self._on_connect
        self._client.on_message = self._on_message
        self._client.on_disconnect = self._on_disconnect

        self._client.connect(self.host, self.port, 60)
        self._client.loop_forever(retry_first_connection=True)


# For backwards compat
MQTTPull = Subscribe



"""Utility classes for the building block 'selector'."""

from typing import List, Callable, Dict, Any, Iterable

from typeguard import typechecked

from pnp import validator
from pnp.models import UDFModel
from pnp.typing import SelectorExpression, Payload
from pnp.utils import Singleton, safe_eval, FallbackBox, EvaluationError


class PayloadSelector(Singleton):
    """
    The actual selector implementation.
    Knows about globals, the suppress literal, udfs and how to evaluate
    simple and complex selector expressions.
    """

    def __init__(self: 'PayloadSelector') -> None:  # pylint: disable=super-init-not-called
        self._suppress_literal = object()
        self._custom = {}  # type: Dict[str, Callable[..., Any]]
        self._register_globals()

    @property
    def suppress(self) -> Any:
        """Return the suppress literal."""
        return self._suppress_literal

    @property
    def suppress_aliases(self) -> List[str]:
        """Returns available aliases for the suppress literal."""
        uppers = ["SUPPRESS", "SUPPRESSME", "SUPPRESSPUSH", "SUPPRESS_ME", "SUPPRESS_PUSH"]
        return uppers + [item.lower() for item in uppers]

    def should_suppress(self, payload: Payload) -> bool:
        """Returns True if the payload is a suppress literal; otherwise False."""
        return payload is self.suppress

    def _register_globals(self) -> None:
        self._custom["abs"] = abs
        self._custom["bool"] = bool
        self._custom["dict"] = dict
        self._custom["float"] = float
        self._custom["getattr"] = getattr
        self._custom["hasattr"] = hasattr
        self._custom["hash"] = hash
        self._custom["int"] = int
        self._custom["isinstance"] = isinstance
        self._custom["len"] = len
        self._custom["list"] = list
        self._custom["max"] = max
        self._custom["min"] = min
        self._custom["ord"] = ord
        self._custom["pow"] = pow
        self._custom["reversed"] = reversed
        self._custom["round"] = round
        self._custom["sorted"] = sorted
        self._custom["str"] = str
        self._custom["zip"] = zip

        from base64 import b64encode, b64decode
        self._custom["b64encode"] = b64encode
        self._custom["b64decode"] = b64decode
        from datetime import datetime
        self._custom["now"] = datetime.now
        self._custom["utcnow"] = datetime.utcnow
        import os
        self._custom["basename"] = os.path.basename

        from .utils import on_off
        self._custom["on_off"] = on_off

    @typechecked
    def register_custom_global(self, name: str, fun: Callable[..., Any]) -> None:
        """
        Register a custom function that should be available in the context of the selector.

        Args:
            name: Name of the function to register.
            fun: The actual function to register.

        Returns:
            None.
        """
        if name not in self._custom:  # Overriding a already existing custom will silently fail!
            self._custom[name] = fun

    def register_udfs(self, udfs: Iterable[UDFModel]) -> None:
        """Register the given user-definied function."""
        validator.all_items(UDFModel, udfs=udfs)
        for udf in udfs:
            self.register_custom_global(udf.name, udf.callable)

    def _eval_wrapper(self, selector: str, payload: Payload) -> Payload:
        suppress_kwargs = {alias: self.suppress for alias in self.suppress_aliases}
        return safe_eval(
            source=selector,
            payload=payload,
            data=payload,
            **suppress_kwargs,
            **self._custom
        )

    @staticmethod
    def _isalambda(v: Any) -> bool:
        lambda_ = lambda: 0
        return isinstance(v, type(lambda_)) and v.__name__ == lambda_.__name__

    def _e(self, snippet: SelectorExpression, payload: Payload) -> Payload:
        # There might be nested dicts or lists inside the complex structure... Eval them recursively
        if isinstance(snippet, dict):
            return {self._e(k, payload): self._e(v, payload) for k, v in snippet.items()}
        if isinstance(snippet, list):
            return [self._e(i, payload) for i in snippet]

        # Test if the snippet constructs a lambda
        # -> assumes to be callable code with payload argument
        try:
            possible_fun = self._eval_wrapper(str(snippet), payload)
        except EvaluationError as eve:
            if str(snippet).startswith('lambda'):
                raise EvaluationError("Your lambda is errorneous: '{snippet}'".format(
                    **locals())) from eve
            # Evaluation failed -> assume that is string literal (dict key/value or list item)
            return snippet

        if not self._isalambda(possible_fun):
            return snippet  # Cannot be executed. So assume it is not a selector expression

        # It is a callable. Lets modify globals and call it :-)
        suppress_kwargs = {alias: self.suppress for alias in self.suppress_aliases}
        customs = self._custom
        for k, v in {**suppress_kwargs, **customs}.items():
            possible_fun.__globals__[k] = v
        try:
            return possible_fun(payload)
        except Exception as exc:
            raise EvaluationError("Error when running the selector lambda: '{snippet}'".format(
                **locals())) from exc

    def eval_selector(self, selector: SelectorExpression, payload: Payload) -> Payload:
        """Applies the specified selector to the given payload."""
        # Wrap payload inside a Box -> this makes dot accessable dictionaries possible
        # We create a dictionary cause payload might not be an actual dictionary.
        # This way wrapping with Box will always work ;-)
        payload = FallbackBox({'base': payload}).base
        if selector is None:
            return payload

        if isinstance(selector, (list, dict)):
            return self._e(selector, payload)  # Complex. Need additional magic

        # No complex structure. We assume that is an expression and we try to evaluate it
        return self._eval_wrapper(str(selector), payload)



"""Contains api endpoints."""

from .base import Endpoint
from .catchall_route import CatchAllRoute, CatchAllRequest
from .health import Health
from .log_level import SetLogLevel
from .metrics import PrometheusExporter
from .ping import Ping
from .trigger import Trigger
from .version import Version

__all__ = [
    'CatchAllRoute',
    'CatchAllRequest',
    'Endpoint',
    'Health',
    'PrometheusExporter',
    'Ping',
    'SetLogLevel',
    'Trigger',
    'Version'
]



"""Base implementation for asynchronous engines."""

import asyncio
from typing import Optional

from pnp.engines._base import Engine, RetryHandler, SimpleRetryHandler, PushExecutor
from pnp.models import TaskSet, TaskModel, PushModel
from pnp.plugins.pull import SyncPull
from pnp.plugins.push import Push
from pnp.shared.async_ import async_sleep_until_interrupt
from pnp.typing import Payload
from pnp.utils import PY37


class AsyncEngine(Engine):
    """Asynchronous engine using asyncio."""

    __REPR_FIELDS__ = 'retry_handler'

    HEARTBEAT_INTERVAL = 0.5

    def __init__(
            self, retry_handler: Optional[RetryHandler] = None
    ):
        super().__init__()
        if not retry_handler:
            self.retry_handler = SimpleRetryHandler()  # type: RetryHandler
        else:
            self.retry_handler = retry_handler
        self.loop = asyncio.get_event_loop()

    async def _start(self, tasks: TaskSet) -> None:
        # Use the loop to create callbacks that was used to start the engine
        self.loop = asyncio.get_event_loop()
        coros = [self._wait_for_tasks_to_complete()]
        for _, task in tasks.items():
            coros.append(self._start_task(task))

        for coro in coros:
            self.loop.create_task(coro)

    async def _stop(self) -> None:
        if not self.tasks:
            return  # Nothing to stop

        await asyncio.gather(
            *[self._stop_task(task) for task in self.tasks.values()]
        )
        await self._wait_for_tasks_to_complete(True)

    async def _wait_for_tasks_to_complete(self, called_from_stop: bool = False) -> None:
        """Check if something is still running on the event loop (like running pushes) so that the
        event loop will not terminate but wait for pending tasks."""
        # pylint: disable=no-member
        fun_pending_tasks = asyncio.all_tasks if PY37 else asyncio.Task.all_tasks  # type: ignore
        fun_current_task = (
            asyncio.current_task if PY37  # type: ignore
            else asyncio.Task.current_task
        )

        # pylint: enable=no-member
        async def _pending_tasks_exist() -> bool:
            all_tasks = list(fun_pending_tasks())
            for task in all_tasks:
                if task is fun_current_task():  # Just in case
                    continue
                if task.done():  # Just in case
                    continue
                # Is a push running?
                if "coro=<AsyncEngine._schedule_push()" in str(task):
                    return True
                # Is a pull running?
                if "coro=<AsyncEngine._start_task()" in str(task):
                    return True
            return False

        await asyncio.sleep(self.HEARTBEAT_INTERVAL)
        while await _pending_tasks_exist():
            await asyncio.sleep(self.HEARTBEAT_INTERVAL)

        if not called_from_stop:
            await self.stop()

    async def _start_task(self, task: TaskModel) -> None:
        """Start the given task."""
        def on_payload_sync(pull: SyncPull, payload: Payload) -> None:
            _ = pull  # Fake usage
            for push in task.pushes:
                self.logger.debug(
                    "[Task-%s] Execution of item '%s' for push '%s'",
                    task.name,
                    payload,
                    push
                )
                self.loop.create_task(self._schedule_push(payload, push))

        task.pull.instance.callback(on_payload_sync)  # type: ignore

        while not task.pull.instance.stopped:
            try:
                await task.pull.instance.pull()

                if not task.pull.instance.stopped and not task.pull.instance.can_exit:
                    # Bad thing... Pulling exited unexpectedly
                    self.logger.error(
                        "[Task-%s] Pulling of '%s' exited unexpectedly",
                        task.name,
                        task.pull.instance,
                    )
            except KeyboardInterrupt:  # pragma: no cover
                self.logger.debug(
                    "[Task-%s] Pulling of '%s' hit a keyboard interrupt",
                    task.name,
                    task.pull.instance
                )
                await self._stop_task(task)
                # self.loop.call_soon_threadsafe(task.pull.instance.stop)
            except:  # pylint: disable=bare-except
                # Bad thing... Pulling exited with exception
                self.logger.exception(
                    "[Task-%s] Pulling of '%s' raised an error",
                    task.name,
                    task.pull.instance.name
                )
            finally:
                await self._handle_pull_exit(task)

    async def _handle_pull_exit(self, task: TaskModel) -> None:
        if task.pull.instance.stopped:
            return
        if task.pull.instance.can_exit:
            await task.pull.instance.stop()
            return

        directive = await self.retry_handler.handle_error()
        if directive.abort:
            self.logger.error(
                "[Task-%s] Pulling of '%s' exited due to retry limitation",
                task.name,
                task.pull.instance,
            )
            await task.pull.instance.stop()
            return

        self.logger.info(
            "[Task-%s] Pulling of '%s' will restart in %s seconds",
            task.name,
            task.pull.instance,
            directive.wait_for
        )

        async def _interrupt() -> bool:
            return task.pull.instance.stopped
        await async_sleep_until_interrupt(directive.wait_for, _interrupt)

    async def _stop_task(self, task: TaskModel) -> None:
        self.logger.info("Stopping task %s", task.name)
        instance = task.pull.instance
        await instance.stop()

    async def _schedule_push(self, payload: Payload, push: PushModel) -> None:
        assert isinstance(push, PushModel)
        assert isinstance(push.instance, Push)

        def _callback(result: Payload, dependency: PushModel) -> None:
            self.loop.create_task(self._schedule_push(result, dependency))

        try:
            await PushExecutor().execute(
                push.instance.name,
                payload,
                push,
                _callback
            )
        except KeyboardInterrupt:  # pragma: no cover
            pass
        except Exception:  # pragma: no cover, pylint: disable=broad-except
            self.logger.exception("Push '%s' failed", push.instance.name)



"""Pull 'n' Push."""
import warnings

__version__ = "0.28.0"
__author__ = "Dennis Muth <d.muth@gmx.net>"


# https://stackoverflow.com/questions/40845304/  \
# runtimewarning-numpy-dtype-size-changed-may-indicate-binary-incompatibility
warnings.filterwarnings("ignore", message=".*numpy.dtype size changed.*")
warnings.filterwarnings("ignore", message=".*numpy.ufunc size changed.*")

warnings.filterwarnings("ignore", message=".*write permissions assigned to anonymous user.*")



"""Http related push plugins."""

import json

import requests

from pnp import utils, validator
from pnp.plugins.push import SyncPush, PushExecutionError, enveloped, parse_envelope


class Call(SyncPush):
    """
    Makes a request to a http resource.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/http.Call/index.md

    """
    __REPR_FIELDS__ = ['fail_on_error', 'method', 'provide_response', 'url']

    def __init__(self, url, method='GET', fail_on_error=False, provide_response=False, **kwargs):
        super().__init__(**kwargs)
        self.url = self._parse_url(url)
        self.method = self._parse_method(method)
        self.fail_on_error = self._parse_fail_on_error(fail_on_error)
        self.provide_response = bool(provide_response)

    @staticmethod
    def _parse_url(val):
        return str(val)

    @staticmethod
    def _parse_method(val):
        val = str(val).upper()
        validator.one_of(utils.HTTP_METHODS, method=val)
        return val

    @staticmethod
    def _parse_fail_on_error(val):
        return utils.try_parse_bool(val)

    @enveloped
    @parse_envelope('url')
    @parse_envelope('method')
    @parse_envelope('fail_on_error')
    def _push(self, url, method, fail_on_error, envelope, payload):  # pylint: disable=arguments-differ
        if isinstance(payload, (dict, list, tuple)):
            try:
                payload = json.dumps(payload)
            except:  # pylint: disable=bare-except
                pass
        resp = requests.request(method, url, data=str(payload))
        if fail_on_error and not 200 <= resp.status_code <= 299:
            raise PushExecutionError(
                "{method} of '{url}' failed with status code = '{status_code}'".format(
                    method=method,
                    url=url,
                    status_code=resp.status_code
                )
            )
        if not self.provide_response:
            return {'data': payload, **envelope} if envelope else payload

        try:
            return dict(status_code=resp.status_code, is_json=True, data=resp.json())
        except ValueError:
            # No valid json, try text
            return dict(status_code=resp.status_code, is_json=False, data=resp.text)



"""Http related plugins."""

import asyncio
from typing import Union, Iterable, Any

from pnp import validator
from pnp.api import RestAPI, APINotConfiguredError
from pnp.api.endpoints import CatchAllRoute, CatchAllRequest
from pnp.plugins.pull import AsyncPull
from pnp.utils import make_list, HTTP_METHODS


class Server(AsyncPull):
    """
    Creates a specific route on the builtin api server and listens to any call to
    that route.
    Any data passed to the endpoint will be tried to be parsed to a dictionary (json).
    If this is not possible the data will be passed as is.

    You need to enable the api via configuration to make this work.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/http.Server/index.md
    """
    __REPR_FIELDS__ = ['allowed_methods', 'prefix_path']

    def __init__(
            self, prefix_path: str, allowed_methods: Union[str, Iterable[str]] = 'GET',
            **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        self.prefix_path = str(prefix_path)
        self.allowed_methods = [str(m).upper() for m in make_list(allowed_methods)]
        validator.subset_of(HTTP_METHODS, allowed_methods=self.allowed_methods)

    async def _incoming(self, request: CatchAllRequest):
        def _make_clean(item):
            if not item:  # Empty string or empty list
                return None
            if not isinstance(item, list):
                return item
            # item -> list
            if len(item) == 1:
                return item[0] if item[0] else None  # Empty string -> None
            return [x for x in item if x]

        query_params = {}
        for param_key, param_value in request.query_params.items():
            value = _make_clean(param_value)
            if value:
                query_params[param_key] = value

        payload = dict(
            endpoint=request.endpoint,
            levels=request.levels,
            method=request.method,
            query=query_params,
            data=request.data,
            is_json=request.is_json_data,
            url=request.url,
            full_path=request.full_path,
            path=request.path
        )
        self.notify(payload)

    async def _pull(self) -> None:
        restapi = RestAPI()
        if not restapi.enabled:
            raise APINotConfiguredError()
        CatchAllRoute(
            route_prefix=self.prefix_path,
            allowed_methods=self.allowed_methods,
            callback=self._incoming
        ).attach(restapi.fastapi)

        while not self.stopped:
            await asyncio.sleep(0.1)



"""Shared functionality for gpio related stuff."""

import logging
import re
from abc import abstractmethod
from collections import defaultdict
from functools import partial
from typing import Optional, Iterable, Any, Callable, Dict, Tuple, List, Union

from pnp.typing import DurationLiteral
from pnp.utils import (
    try_parse_int, Debounce, parse_duration_literal, Singleton,
    Loggable
)

_LOGGER = logging.getLogger(__name__)

CONST_AVAILABLE_GPIO_PINS = list(range(2, 28))
CONST_RISING_OPTIONS = ["rise", "rising"]
CONST_FALLING_OPTIONS = ["fall", "falling"]
CONST_SWITCH_OPTIONS = ["button", "switch"]
CONST_MOTION_OPTIONS = ["motion"]

CONST_RISING = "rising"
CONST_FALLING = "falling"
CONST_SWITCH = "switch"
CONST_MOTION_ON = "motion_on"
CONST_MOTION_OFF = "motion_off"

Channel = int
Edge = str
Bounce = int
EventCallback = Callable[[Channel], None]
InterceptCallback = Callable[[Channel, Edge], None]
CallbackDict = Dict[Channel, List[Tuple[Edge, EventCallback, Optional[int]]]]


class GPIOAdapter(Singleton, Loggable):
    """Convenience wrapper around the RPi.GPIO package."""

    def __init__(self) -> None:
        super().__init__()
        self.logger.debug("GPIOWrapper __init__ called")
        self.GPIO = self._load_gpio_package()
        self.GPIO.setmode(self.GPIO.BCM)
        self._mchannel = dict()  # type: Dict[Channel, Optional[Edge]]
        self._callbacks = defaultdict(list)  # type: CallbackDict

    def add_event_detect(
        self, channel: Channel, edge: Edge, callback: EventCallback,
        bouncetime: Optional[Bounce] = None
    ) -> None:
        """Adds a callback when the state of the channel changes."""
        self._callbacks[channel].append((edge, callback, bouncetime))

    def apply(self) -> None:
        """Applies the configured callbacks."""
        def safe_max(lst: Iterable[Bounce]) -> Optional[Bounce]:
            if not lst:
                return None
            return max(lst)

        for channel, event in self._callbacks.items():
            self.GPIO.setup(channel, self.GPIO.IN)
            # If multiple bounce times are specified, use max
            max_bouncetime = safe_max([bouncetime for _, _, bouncetime in event if bouncetime])
            edges = list({edge for edge, _, _ in event})
            edge = edges[0] if len(edges) == 1 else self.GPIO.BOTH

            # We have to use both mode because the pin is configured to use rising and falling.
            # But the callback does not provide the edge, so we have to guess in those cases ;-)
            if edge == self.GPIO.BOTH:
                self.logger.debug(
                    "Channel '%s' does provide rising and falling - Using 'both'", channel
                )
                self._mchannel[channel] = None

            gpio_edge = self._str_to_gpio_mode(edge)
            if max_bouncetime:
                self.logger.debug("Channel '%s' uses bouncetime of %sms. "
                                  "This will apply for all callbacks", channel, max_bouncetime)
                self.GPIO.add_event_detect(channel, gpio_edge, bouncetime=max_bouncetime)
            else:
                self.GPIO.add_event_detect(channel, gpio_edge)

            self.GPIO.add_event_callback(channel, self._dispatcher)

    def remove_event_detect(self, channel: Channel) -> None:
        """Removes a previously added callback."""
        self._mchannel.pop(channel, None)
        self._callbacks.pop(channel, None)
        self.GPIO.remove_event_detect(channel)

    def cleanup(self) -> None:
        """Clean up the mess you have done."""
        self._mchannel.clear()
        self._callbacks.clear()
        self.GPIO.cleanup()

    def _dispatcher(self, channel: Channel) -> None:
        callbacks = self._callbacks.get(channel, None)
        if not callbacks:
            return
        is_multi_channel = channel in self._mchannel
        current_edge = None
        if is_multi_channel:
            last_state = self._mchannel[channel] or CONST_FALLING
            current_edge = CONST_RISING if last_state == CONST_FALLING else CONST_FALLING
            self._mchannel[channel] = current_edge

        for cb_edge, cback, _ in callbacks:
            if current_edge is None or cb_edge == current_edge:
                cback(channel)

    def _str_to_gpio_mode(self, _str: str) -> Edge:
        if _str == self.GPIO.BOTH:
            return _str
        _str = str(_str)
        if _str.lower() in CONST_RISING_OPTIONS:
            return self.GPIO.RISING  # type: ignore
        if _str.lower() in CONST_FALLING_OPTIONS:
            return self.GPIO.FALLING  # type: ignore
        raise ValueError("The given string '{}' is not a valid GPIO mode".format(_str))

    @staticmethod
    def _load_gpio_package() -> Any:
        try:
            try:
                import RPi.GPIO as GPIO
            except ImportError:
                from RPi import GPIO  # Only works for test cases when using the mocked package
        except ImportError:  # pragma: no cover
            _LOGGER.warning("RPi.GPIO package is not available - Using mock")
            from ..mocking import GPIOMock as GPIO  # type: ignore
        return GPIO


class Callback:
    """Base class for a gpio callback."""

    __REPR_FIELDS__ = 'gpio_pin'

    def __init__(self, gpio_pin: str):
        self.gpio_pin = self._str_to_gpio_pin(gpio_pin)

    @property
    def GPIO(self) -> GPIOAdapter:
        """Return the gpio wrapper."""
        return GPIOAdapter()

    @classmethod
    def load(cls, gpio_pin_str: str, mode_str: str) -> 'Callback':
        """Make an instance using the gpio_pin as a string. Override in child classes
        to perform some validation checks."""
        raise NotImplementedError()

    @classmethod
    def from_str(cls, candidate: Any, default: str = CONST_RISING) -> 'Callback':
        """Load the callback from a string."""
        gpio_pin_str, _, mode_str = str(candidate).partition(':')
        if not mode_str:
            mode_str = default
        if any(mode_str.lower().startswith(aswitch) for aswitch in CONST_RISING_OPTIONS):
            return RisingCallback.load(gpio_pin_str, mode_str)
        if any(mode_str.lower().startswith(aswitch) for aswitch in CONST_FALLING_OPTIONS):
            return FallingCallback.load(gpio_pin_str, mode_str)
        if any(mode_str.lower().startswith(aswitch) for aswitch in CONST_SWITCH_OPTIONS):
            return SwitchCallback.load(gpio_pin_str, mode_str)
        if any(mode_str.lower().startswith(aswitch) for aswitch in CONST_MOTION_OPTIONS):
            return MotionCallback.load(gpio_pin_str, mode_str)
        raise ValueError("The specified pin mode '{}' is not supported".format(mode_str))

    def _intercept(self, channel: Channel, event: Edge, callback: InterceptCallback) -> None:
        _ = self  # Fake usage
        callback(channel, event)

    @abstractmethod
    def run(self, callback: InterceptCallback) -> None:
        """Registers this callback to the gpio wrapper."""
        raise NotImplementedError()  # pragma: no cover

    def stop(self) -> None:
        """Stop this callback to receive events."""
        self.GPIO.remove_event_detect(self.gpio_pin)

    def _eqstr(self) -> str:
        # If we have multiple switches or motion for the same pin only one will applied
        # We basically account for the pin and the class only without additional params.
        return "{}({})".format(type(self).__name__, str(self.gpio_pin))

    @staticmethod
    def _parse_mode_str(mode_str: str, regex: str, *groups: str) -> Iterable[str]:
        _rgx = re.compile(regex)
        _match = _rgx.match(mode_str)
        if not _match:
            raise ValueError("The given mode '{}' is not a valid definition for a callback"
                             .format(mode_str))
        return [_match.group(g) for g in groups]

    @staticmethod
    def _str_to_gpio_pin(_str: str) -> Channel:
        pin = try_parse_int(_str)
        if not pin or pin not in CONST_AVAILABLE_GPIO_PINS:
            lmin = min(CONST_AVAILABLE_GPIO_PINS)
            lmax = max(CONST_AVAILABLE_GPIO_PINS)
            raise ValueError("The given string '{}' is not a valid GPIO pin. "
                             "String needs to be convertible to int and between {} and {}"
                             .format(_str, lmin, lmax))
        return pin

    def __hash__(self) -> int:
        return hash(str(self._eqstr()))

    def __eq__(self, other: Any) -> bool:
        if not hasattr(other, '_eqstr'):  # pragma: no cover
            return False
        return bool(self._eqstr() == other._eqstr())  # pylint: disable=protected-access


class RisingCallback(Callback):
    """Callback when the gpio channel switches to high."""
    EDGE = CONST_RISING

    @classmethod
    def load(cls, gpio_pin_str: str, mode_str: str) -> 'Callback':
        """Make an instance using the gpio_pin as a string."""
        return cls(gpio_pin_str)

    def run(self, callback: InterceptCallback) -> None:
        self.GPIO.add_event_detect(
            self.gpio_pin,
            self.EDGE,
            callback=partial(self._intercept, event=self.EDGE, callback=callback)
        )


class FallingCallback(RisingCallback):
    """Callback when the gpio channel switches to low."""
    EDGE = CONST_FALLING


class SwitchCallback(Callback):
    """Callback when the gpio channel switches to high with some bounce delay."""
    EDGE = CONST_RISING
    BOUNCE_REGEX = r"^(switch|button)(\((?P<delay>\d+)\))?$"
    BOUNCE_GROUP = "delay"
    BOUNCE_DEFAULT = 500

    def __init__(self, gpio_pin: str, delay: Union[Bounce, DurationLiteral]):
        super().__init__(gpio_pin)
        self.delay = int(delay)

    @classmethod
    def load(cls, gpio_pin_str: str, mode_str: str) -> 'Callback':
        """Make an instance using the gpio_pin as a string."""
        delay, = cls._parse_mode_str(mode_str, cls.BOUNCE_REGEX, cls.BOUNCE_GROUP)
        return cls(gpio_pin_str, delay or cls.BOUNCE_DEFAULT)

    def run(self, callback: InterceptCallback) -> None:
        self.GPIO.add_event_detect(
            self.gpio_pin,
            self.EDGE,
            bouncetime=self.delay,
            callback=partial(self._intercept, event=CONST_SWITCH, callback=callback)
        )


class MotionCallback(RisingCallback):
    """Callback when the gpio channel switches to high. When the gpio channel switches back to low
    the callback will emit an motion off event after some predefined delay."""
    DELAY_REGEX = r"^motion(\((?P<delay>\d+[sSmMhHdDwW]?)\))?$"
    DELAY_GROUP = "delay"
    DELAY_DEFAULT = "30s"

    def __init__(self, gpio_pin: str, delay: str):
        super().__init__(gpio_pin)
        self.delay = parse_duration_literal(delay)
        self._debouncer = None  # type: Optional[Debounce]

    @classmethod
    def load(cls, gpio_pin_str: str, mode_str: str) -> 'Callback':
        """Make an instance using the gpio_pin as a string."""
        delay, = cls._parse_mode_str(mode_str, cls.DELAY_REGEX, cls.DELAY_GROUP)
        return cls(gpio_pin_str, delay or cls.DELAY_DEFAULT)

    def _motion_off(self, channel: Channel, callback: InterceptCallback) -> None:
        callback(channel, CONST_MOTION_OFF)
        self._debouncer = None

    def _intercept(self, channel: Channel, event: Edge, callback: InterceptCallback) -> None:
        if event == CONST_RISING:
            if self._debouncer is None:
                callback(channel, CONST_MOTION_ON)
                self._debouncer = Debounce(self._motion_off, self.delay)
            self._debouncer(channel, callback)

    def stop(self) -> None:
        if self._debouncer:
            self._debouncer.execute_now()



"""Contains base classes for config parsers."""
from abc import abstractmethod
from typing import Optional, Iterable, Any, List

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from pnp.engines import Engine
from pnp.models import TaskSet, UDFModel, PullModel, APIModel


class Configuration(BaseModel):
    """Represents a parsed, instantiated and valid configuration."""

    # API instance if configured or None
    api: Optional[APIModel]

    # A set of tasks of this configuration
    tasks: TaskSet

    # The engine that is configured; might be None if no engine was specified
    engine: Optional[Engine]

    # The configured udfs. List is empty if no udf is configured
    udfs: List[UDFModel]

    class Config:
        """Pydantic configuration."""
        arbitrary_types_allowed = True


class ConfigLoader:
    """Base class / interface for any configuration loader."""

    @classmethod
    def supported_extensions(cls) -> Iterable[str]:
        """Returns an iterable with all supported file extensions."""
        raise NotImplementedError()

    @abstractmethod
    def load_pull_from_snippet(self, snippet: Any, name: str, **extra: Any) -> PullModel:
        """Return a new pull model instantiated from a config snippet.

        Args:
            snippet (Any): The configuration snippet to configure a pull.
            name (str): The name of the pull (if part of the snippet; the config will take
              precedence)
            extra: Any extra arguments to instantiate the pull.

        Return:
            The instantiated pull model.
        """
        raise NotImplementedError()

    @abstractmethod
    def load_config(self, config_file: str) -> Configuration:
        """Load the configuration from the specified config file.

        Args:
            config_file: The config file to load. Which files are supported depends on the loader.

        Return:
            A full-blown / instantiated configuration.
        """
        raise NotImplementedError()



"""Monitoring related plugins."""

import os
import subprocess

import psutil

from pnp.plugins.pull import SyncPolling


class Stats(SyncPolling):
    """
    Periodically emits stats about the host system, like cpu_use, memory_use, swap_use, ...

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/monitor.Stats/index.md
    """

    @staticmethod
    def _root_path():
        return os.path.abspath(os.sep)

    @classmethod
    def _disk_usage(cls):
        try:
            return psutil.disk_usage(cls._root_path()).percent
        except FileNotFoundError:  # pragma: no cover
            return 0.0

    @staticmethod
    def _cpu_temp():
        try:
            # Raspberry
            with open('/sys/class/thermal/thermal_zone0/temp') as tzone:
                return round(float(tzone.read()) * 0.001, 1)
        except FileNotFoundError:  # pragma: no cover
            return 0.0

    @staticmethod
    def _cpu_freq():
        freq = psutil.cpu_freq()
        if not freq:
            return 0
        return freq.current

    @staticmethod
    def _throttled():
        # Raspberry only
        try:
            mask1 = 1  # under voltage
            mask2 = 2  # arm frequency capped
            mask4 = 4  # throttled
            mask8 = 8  # soft temperature limit throttle (pi3 b+)

            out = subprocess.run(
                ['vcgencmd', 'get_throttled'],
                stdout=subprocess.PIPE,
                check=False
            ).stdout.decode('utf-8')
            hex_num = int(out[out.find('=') + 1:], 16)

            return (
                int(bool(hex_num & mask1)),
                int(bool(hex_num & mask2)),
                int(bool(hex_num & mask4)),
                int(bool(hex_num & mask8)),
            )
        except FileNotFoundError:
            return 0, 0, 0, 0

    def _poll(self):
        l1m, l5m, l15m = os.getloadavg()
        uvolt, fcap, throttled, temp_limit = self._throttled()
        return {
            'cpu_count': psutil.cpu_count(),
            'cpu_freq': self._cpu_freq(),
            'cpu_temp': self._cpu_temp(),
            'cpu_use': psutil.cpu_percent(),
            'disk_use': self._disk_usage(),
            'load_1m': l1m,
            'load_5m': l5m,
            'load_15m': l15m,
            'memory_use': psutil.virtual_memory().percent,
            'rpi_cpu_freq_capped': fcap,
            'rpi_temp_limit_throttle': temp_limit,
            'rpi_throttle': throttled,
            'rpi_under_voltage': uvolt,
            'swap_use': psutil.swap_memory().percent
        }



"""FTP related pulls."""

import copy
import tempfile
from contextlib import contextmanager

from pnp import validator
from pnp.plugins import load_optional_module
from pnp.plugins.pull import SyncPull
from pnp.utils import is_iterable_but_no_str, make_list


class Server(SyncPull):
    """
    Runs a ftp server on the specified port to receive and send files by ftp protocol.
    Optionally sets up a simple user/password authentication mechanism.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/ftp.Server/index.md
    """
    __REPR_FIELDS__ = ['directory', 'events', 'max_cons', 'max_cons_ip', 'password', 'port', 'user']

    EXTRA = 'ftp'

    EVENT_CONNECT = 'connect'
    EVENT_DISCONNECT = 'disconnect'
    EVENT_FILE_RECEIVED = 'file_received'
    EVENT_FILE_RECEIVED_INCOMPLETE = 'file_received_incomplete'
    EVENT_FILE_SENT = 'file_sent'
    EVENT_FILE_SENT_INCOMPLETE = 'file_sent_incomplete'
    EVENT_LOGIN = 'login'
    EVENT_LOGOUT = 'logout'

    ALL_EVENTS = [EVENT_CONNECT, EVENT_DISCONNECT, EVENT_FILE_RECEIVED,
                  EVENT_FILE_RECEIVED_INCOMPLETE, EVENT_FILE_SENT, EVENT_FILE_SENT_INCOMPLETE,
                  EVENT_LOGIN, EVENT_LOGOUT]

    def __init__(self, directory=None, port=2121, user_pwd=None, events=None,
                 max_cons=256, max_cons_ip=5, **kwargs):
        super().__init__(**kwargs)
        self.port = int(port)
        self.max_cons = int(max_cons)
        self.max_cons_ip = int(max_cons_ip)
        validator.is_instance(str, allow_none=True, directory=directory)
        if directory:
            validator.is_directory(directory=directory)
        self.directory = directory
        if not user_pwd:  # No password -> anonymous access
            self.user = None
        elif isinstance(user_pwd, str):
            self.user = user_pwd
            self.password = ''
        elif is_iterable_but_no_str(user_pwd) and user_pwd:  # at least one element
            self.user = user_pwd[0]
            if len(user_pwd) > 1:  # Password too
                self.password = user_pwd[1]
        else:
            raise TypeError("Argument 'user_pwd' is expected to be a str (user) or a tuple of "
                            "user and password.")

        self.events = make_list(events) or copy.copy(self.ALL_EVENTS)
        if self.events:
            validator.subset_of(self.ALL_EVENTS, events=self.events)

        self.server = None

    def _pull(self):
        with self._setup_directory() as ftp_directory:
            authorizer = self._setup_auth(ftp_directory)
            handler = self._setup_handler(authorizer)
            self.server = self._setup_server(handler)

            try:
                self.server.serve_forever()
            except OSError:
                # Bug on OSX when closing: [Errno 9] Bad file descriptor
                import traceback
                self.logger.warning("%s", traceback.format_exc())

    def _stop(self):
        if self.server:
            self.server.close_all()
        return super()._stop()

    @contextmanager
    def _setup_directory(self):
        if self.directory:
            yield self.directory
        else:
            with tempfile.TemporaryDirectory() as tmpdir:
                yield tmpdir

    def _setup_auth(self, ftp_directory):
        auth = load_optional_module('pyftpdlib.authorizers', self.EXTRA)
        authorizer = auth.DummyAuthorizer()
        if self.user:
            self.logger.debug(
                "Setting up authorization: %s:%s @ %s",
                self.user, '*' * len(self.password), ftp_directory
            )
            authorizer.add_user(self.user, self.password, ftp_directory, perm='elradfmw')
        else:
            self.logger.debug("Setting up anonymous access @ %s", ftp_directory)
            authorizer.add_anonymous(ftp_directory, perm='elradfmw')

        return authorizer

    def _setup_handler(self, authorizer):
        handlers = load_optional_module('pyftpdlib.handlers', self.EXTRA)
        handler = handlers.FTPHandler
        handler.authorizer = authorizer
        handler.banner = 'Welcome to pnp embedded ftp server'
        self._register_callbacks(handler)
        return handler

    def _setup_server(self, handler):
        self.logger.debug("Setting up server on 127.0.0.1:%s", str(self.port))
        servers = load_optional_module('pyftpdlib.servers', self.EXTRA)
        address = ('', self.port)
        server = servers.ThreadedFTPServer(address, handler)
        server.max_cons = self.max_cons
        server.max_cons_per_ip = self.max_cons_ip
        return server

    def _register_callbacks(self, handler):
        if self.EVENT_CONNECT in self.events:
            handler.on_connect = lambda that: \
                self.notify({'event': self.EVENT_CONNECT, 'data': {}})
        if self.EVENT_DISCONNECT in self.events:
            handler.on_disconnect = lambda that: \
                self.notify({'event': self.EVENT_DISCONNECT, 'data': {}})
        if self.EVENT_LOGIN in self.events:
            handler.on_login = lambda that, user: \
                self.notify({'event': self.EVENT_LOGIN, 'data': {'user': user}})
        if self.EVENT_LOGOUT in self.events:
            handler.on_logout = lambda that, user: \
                self.notify({'event': self.EVENT_LOGOUT, 'data': {'user': user}})
        if self.EVENT_FILE_SENT in self.events:
            handler.on_file_sent = lambda that, file_path: \
                self.notify({'event': self.EVENT_FILE_SENT, 'data': {'file_path': file_path}})
        if self.EVENT_FILE_RECEIVED in self.events:
            handler.on_file_received = lambda that, file_path: \
                self.notify({'event': self.EVENT_FILE_RECEIVED, 'data': {'file_path': file_path}})
        if self.EVENT_FILE_SENT_INCOMPLETE in self.events:
            handler.on_incomplete_file_sent = lambda that, file_path: \
                self.notify({
                    'event': self.EVENT_FILE_SENT_INCOMPLETE,
                    'data': {'file_path': file_path}
                })
        if self.EVENT_FILE_RECEIVED_INCOMPLETE in self.events:
            handler.on_incomplete_file_received = lambda that, file_path: \
                self.notify({
                    'event': self.EVENT_FILE_RECEIVED_INCOMPLETE,
                    'data': {'file_path': file_path}
                })



"""Basic stuff for implementing pull plugins."""

import asyncio
import inspect
import multiprocessing as proc
from abc import abstractmethod
from datetime import datetime
from typing import Any, Callable, Optional

from schedule import Scheduler  # type: ignore
from typeguard import typechecked

from pnp.plugins import Plugin
from pnp.shared.async_ import (
    async_sleep_until_interrupt,
    run_sync
)
from pnp.typing import Payload
from pnp.utils import (
    parse_duration_literal,
    try_parse_bool,
    DurationLiteral,
    sleep_until_interrupt
)


class PollingError(Exception):
    """Base class for errors that occur during polling."""


class StopPollingError(Exception):
    """Raise this exception in child classes if you want to stop the polling scheduler."""


PullCallback = Callable[['Pull', Payload], None]


class Pull(Plugin):
    """
    Base class for pull plugins.
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._assert_pull_compat()
        self._assert_pull_now_compat()

        self._stopped = proc.Event()
        self._callback: Optional[PullCallback] = None

    @property
    def stopped(self) -> bool:
        """Returns True if the pull is considered stopped; otherwise False."""
        return self._stopped.is_set()

    @property
    def can_exit(self) -> bool:
        """Override this property when an exit of the pull can be expected. This
        will prohibit the engine (or should) from raising notifications / errors
        or countermeasures like retries."""
        return False

    @property
    def supports_pull_now(self) -> bool:
        """Returns True if this instance supports an immediate pull; otherwise False."""
        return isinstance(self, (AsyncPullNowMixin, SyncPullNowMixin))

    @typechecked
    def callback(self, value: PullCallback) -> None:
        """Sets the payload callback on this instance."""
        self._callback = value

    def notify(self, payload: Payload) -> None:
        """Emit some payload to the execution engine."""
        if self._callback:
            self._callback(self, payload)

    def _assert_pull_compat(self) -> None:
        self._assert_abstract_compat((SyncPull, AsyncPull))
        self._assert_fun_compat('_pull')
        self._assert_fun_compat('_stop')

    def _assert_pull_now_compat(self) -> None:
        that = self
        if isinstance(that, (SyncPullNowMixin, AsyncPullNowMixin)):
            self._assert_fun_compat('_pull_now')

    async def pull(self) -> None:
        """Performs the actual data pulling."""
        pull_fun = getattr(self, '_pull')
        if inspect.iscoroutinefunction(pull_fun):
            await pull_fun()
            return
        await run_sync(pull_fun)

    async def pull_now(self) -> None:
        """Performs a pull now. Be careful: Not every pull does support this. Make sure to call
        supports_pull_now() previously to check the compatibility.
        """
        pull_now_fun = getattr(self, '_pull_now')
        if inspect.iscoroutinefunction(pull_now_fun):
            await pull_now_fun()
            return
        await run_sync(pull_now_fun)

    async def stop(self) -> None:
        """Stops the execution of this pull."""
        stop_fun = getattr(self, '_stop')
        if inspect.iscoroutinefunction(stop_fun):
            await stop_fun()
            return
        await run_sync(stop_fun)


class SyncPull(Pull):
    """Base class for synchronous pulls."""

    @abstractmethod
    def _pull(self) -> None:
        """
        Performs the actual data retrieval / task.
        Returns:
            None.
        """
        raise NotImplementedError()  # pragma: no cover

    def _stop(self) -> None:
        """Async variant of `stop()`."""
        self._stopped.set()

    def _sleep(self, sleep_time: float = 10) -> None:
        """Call in subclass to perform some sleeping."""
        sleep_until_interrupt(sleep_time, lambda: self.stopped, interval=0.5)


class AsyncPull(Pull):
    """
    Base class for asynchronous pulls.
    """
    @abstractmethod
    async def _pull(self) -> None:
        """Performs the actual data retrieval in a way that is compatible with the async engine."""
        raise NotImplementedError()

    async def _stop(self) -> None:
        """Async variant of `stop()`."""
        self._stopped.set()

    async def _sleep(self, sleep_time: float = 10) -> None:
        """Call in subclass to perform some sleeping."""
        async def _interrupt() -> bool:
            return self.stopped
        await async_sleep_until_interrupt(sleep_time, _interrupt, interval=0.5)


class SyncPullNowMixin:
    """Adds support to execute the pull right now without waiting for a specific trigger
    (like cron, special events, ...). This comes in handy for api triggers or the RunOnce pull."""

    def _pull_now(self) -> None:
        """Executes the pull right now."""
        raise NotImplementedError()


class AsyncPullNowMixin:
    """Adds support to asynchronously execute the pull right now without waiting for a
    specific trigger (like cron, special events, ...). This comes in handy for api
    triggers or the RunOnce pull."""

    async def _pull_now(self) -> None:
        """Asynchronously executes the pull right now."""
        raise NotImplementedError()


class Polling(AsyncPull, AsyncPullNowMixin):
    """
    Base class for polling plugins.

    You may specify duration literals such as 60 (60 secs), 1m, 1h (...) to realize a periodic
    polling or cron expressions (*/1 * * * * > every min) to realize cron like behaviour.
    """
    __REPR_FIELDS__ = ['interval', 'is_cron']

    def __init__(
        self, interval: Optional[DurationLiteral] = 60, instant_run: bool = False, **kwargs: Any
    ):
        super().__init__(**kwargs)

        self._assert_polling_compat()

        if interval is None:
            # No scheduled execution. Use endpoint `/trigger` of api to execute.
            self._poll_interval = None
            self.interval = None
            self.is_cron = False
        else:
            try:
                # Literals such as 60s, 1m, 1h, ...
                self._poll_interval = parse_duration_literal(interval)
                self.interval = interval
                self.is_cron = False
            except TypeError:
                # ... or a cron-like expression is valid
                from cronex import CronExpression  # type: ignore
                self._cron_interval = CronExpression(interval)
                self.interval = self._cron_interval
                self.is_cron = True

        self._is_running = False
        self._scheduler: Optional[Scheduler] = None
        self._instant_run = try_parse_bool(instant_run, False)

    def _assert_polling_compat(self) -> None:
        self._assert_abstract_compat((SyncPolling, AsyncPolling))
        self._assert_fun_compat('_poll')

    async def _pull(self) -> None:
        def _callback() -> None:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                asyncio.ensure_future(self._run_schedule())

        self._scheduler = Scheduler()
        self._configure_scheduler(self._scheduler, _callback)

        if self._instant_run:
            self._scheduler.run_all()

        while not self.stopped:
            self._scheduler.run_pending()
            await self._sleep(0.5)

        while self._is_running:  # Keep the loop alive until the job is finished
            await asyncio.sleep(0.1)

    async def _pull_now(self) -> None:
        await self._run_now()

    async def _run_now(self) -> Payload:
        """Runs the poll right now. It will not run, if the last poll is still running."""
        if self._is_running:
            self.logger.warning("Polling job is still running. Skipping current run")
            return

        self._is_running = True
        try:
            payload = await self.poll()

            if payload is not None:
                self.notify(payload)

            return payload
        finally:
            self._is_running = False

    async def _run_schedule(self) -> None:
        try:
            if self.is_cron:
                dtime = datetime.now()
                if not self._cron_interval.check_trigger((
                        dtime.year, dtime.month, dtime.day,
                        dtime.hour, dtime.minute
                )):
                    return  # It is not the time for the cron to trigger

            await self._run_now()
        except StopPollingError:
            await self._stop()
        except Exception:  # pragma: no cover, pylint: disable=broad-except
            self.logger.exception("Polling of '%s' failed", self.name)

    def _configure_scheduler(self, scheduler: Scheduler, callback: Callable[[], None]) -> None:
        """
        Configures the scheduler. You have to differ between "normal" intervals and
        cron like expressions by checking `self.is_cron`.

        Override in subclasses to fir the behaviour to your needs.

        Args:
            scheduler (schedule.Scheduler): The actual scheduler.
            callback (callable): The callback to call when the time is right.

        Returns:
            None
        """
        if self.is_cron:
            # Scheduler always executes at the exact minute to check for cron triggering
            scheduler.every().minute.at(":00").do(callback)
        else:
            # Only activate when an interval is specified
            # If not the only way is to trigger the poll by the api `trigger` endpoint
            if self._poll_interval:
                # Scheduler executes every interval seconds to execute the poll
                scheduler.every(self._poll_interval).seconds.do(callback)

    async def poll(self) -> Payload:
        """Performs polling."""
        poll_fun = getattr(self, '_poll')
        if inspect.iscoroutinefunction(poll_fun):
            return await poll_fun()
        return await run_sync(poll_fun)


class SyncPolling(Polling):
    """
    Base class for polling plugins who can poll synchronously.
    """

    @abstractmethod
    def _poll(self) -> Payload:
        """
        Implement in plugin components to do the actual polling.

        Returns: Returns the data for the downstream pipeline.
        """
        raise NotImplementedError()  # pragma: no cover


class AsyncPolling(Polling):
    """
    Base class for polling plugins who can poll asynchronously.
    """

    @abstractmethod
    async def _poll(self) -> Payload:
        """
        Implement in plugin components to do the actual polling.

        Returns: Returns the data for the downstream pipeline.
        """
        raise NotImplementedError()  # pragma: no cover



"""Some mocking stuff for packages that are not available on every platform (just for testing)."""

import random

from collections import defaultdict


class DHTMock:
    """Mocks a DHT device (only available on rpi's)."""
    DHT22 = "dht22"
    DHT11 = "dht11"
    AM2302 = "am2302"

    @staticmethod
    def read_retry(sensor, pin):  # pylint: disable=unused-argument
        """Read the sensor values (humidity, temperature)."""
        return (
            round(random.uniform(1, 100), 2),
            round(random.uniform(8, 36), 2)
        )  # pragma: no cover


class GPIOMock:  # pragma: no cover
    """Mocks the RPi.GPIO (only available on rpi)."""
    BCM = "bcm"
    IN = "in"
    RISING = "rising"
    FALLING = "falling"
    BOTH = "both"

    MODES = dict()
    CALLBACKS = defaultdict(list)
    SETUPS = set()

    @classmethod
    def setmode(cls, mode):
        """Set the mode."""

    @classmethod
    def setup(cls, channel, in_out):  # pylint: disable=unused-argument
        """Setup the given gpio channel."""
        if channel in cls.SETUPS:
            raise RuntimeError(
                "You called setup more than one time for channel '{}'".format(channel)
            )
        cls.SETUPS.add(channel)

    @classmethod
    def cleanup(cls):
        """Free the resources."""

    @classmethod
    def add_event_detect(cls, channel, mode, callback=None, bouncetime=None):  # pylint: disable=unused-argument
        """Adds an event detection callback when the state of the gpio channels changes."""
        if channel not in cls.SETUPS:
            raise RuntimeError(
                "You did not setup the channel '{}' before activating event "
                "detection".format(channel)
            )
        cls.MODES[channel] = mode
        if not callback:
            return
        if channel in cls.CALLBACKS:
            raise RuntimeError("You already added a callback to this channel")
        cls.CALLBACKS[channel].append((callback, mode))

    @classmethod
    def add_event_callback(cls, channel, callback):
        """Adds an event detection callback when the state of the gpio channels changes."""
        if channel not in cls.MODES:
            raise RuntimeError("You have to enable the event detection first")
        mode = cls.MODES[channel]
        cls.CALLBACKS[channel].append((callback, mode))

    @classmethod
    def remove_event_detect(cls, channel):
        """Remove all callbacks from the event detection for the given gpio channel."""
        cls.CALLBACKS.pop(channel, None)
        cls.MODES.pop(channel, None)

    @classmethod
    def clear(cls):
        """Clear this instance for re-use."""
        cls.MODES.clear()
        cls.CALLBACKS.clear()
        cls.SETUPS.clear()

    @classmethod
    def fire_event(cls, channel, mode):
        """Fire an event in the specified mode for the given gpio channel."""
        for cback, cb_mode in cls.CALLBACKS[channel]:
            if cb_mode in (cls.BOTH, mode):
                cback(channel)


class FritzBoxHostsMock:  # pragma: no cover
    """Mocks the fritzconnection.FritzHosts class."""
    def __init__(self):
        self.address = None
        self.user = None
        self.password = None
        self.calls = 0

    def __call__(self, address, user, password):
        """The actual instantiation of mock from a client perspective."""
        self.address = address
        self.user = user
        self.password = password
        return self

    @property
    def modelname(self):
        """Return the name of the fritzbox model / make."""
        return "Fritz!Box Mock v1.0"

    def get_hosts_info(self):
        """Return the known hosts."""
        _ = self  # Fake usage
        self.calls += 1
        return [
            {"mac": "12:34:56:78:12", 'ip': '192.168.178.10', 'name': 'pc1', 'status': True},
            {"mac": "12:34:56:78:13", 'ip': '192.168.178.11', 'name': 'pc2', 'status': False},
            {"mac": "12:34:56:78:14", 'ip': '192.168.178.12', 'name': 'pc3',
             'status': self.calls <= 1}
        ]

    def get_specific_host_entry(self, mac_address):
        """Return the host associated to the given mac address."""
        if mac_address == '12:34:56:78:14':
            self.calls += 1
        _map = {
            '12:34:56:78:12': {
                'NewIPAddress': '192.168.178.10', 'NewActive': True, 'NewHostName': 'pc1'
            },
            '12:34:56:78:13': {
                'NewIPAddress': '192.168.178.11', 'NewActive': False, 'NewHostName': 'pc2'
            },
            '12:34:56:78:14': {
                'NewIPAddress': '192.168.178.12', 'NewActive': self.calls <= 1, 'NewHostName': 'pc3'
            }
        }
        return _map[mac_address]


class PyAudioMock:  # pragma: no cover
    """Mocks the pyaudio package (only available with audio devices)."""
    class _StreamMock:
        def __init__(self, wav_file):
            self.wav_file = wav_file
            self.fhandle = open(self.wav_file, 'rb')

        def read(self, chunk_size, exception_on_overflow):  # pylint: disable=unused-argument
            """Read from the stream."""
            res = self.fhandle.read()
            self.fhandle.seek(0)
            return res

        def close(self):
            """Close the stream."""
            self.fhandle.close()

    def __init__(self, mock_wav):
        self.mock_wav = str(mock_wav)

    def __call__(self):
        return self

    def open(self, **kwargs):  # pylint: disable=unused-argument
        """Open a audio stream."""
        return self._StreamMock(self.mock_wav)

    def terminate(self):
        """Terminate the audio stream."""



"""Contains a trigger endpoint."""

import logging

from fastapi import FastAPI, Query, HTTPException
from starlette.responses import JSONResponse

from pnp.api.models import EmptyResponse
from pnp.models import TaskSet
from .base import Endpoint

_LOGGER = logging.getLogger(__name__)


class Trigger(Endpoint):
    """Triggers a poll right now without being it's schedule be fulfilled.
    This only works for polling components and not for regular pull"""

    def __init__(self, tasks: TaskSet):
        self.tasks = tasks

    async def endpoint(
            self,
            task_name: str = Query(
                ...,
                title="Task",
                description="The name of the task to trigger right now",
                alias="task"
            )
    ) -> JSONResponse:
        """Triggers a poll right now without being it's schedule be fulfilled.
        This only works for polling components and not for regular pull"""

        task = self.tasks.get(task_name)
        if not task:
            raise HTTPException(
                status_code=422,
                detail=f"Given task name '{task_name}' is not a known task."
            )

        pull = task.pull.instance
        if not pull.supports_pull_now:
            raise HTTPException(
                status_code=422,
                detail=f"Task '{task_name}' does not support pull_now(). "
                       f"Implement PullNowMixin / AsyncPullNowMixin for support"
            )

        try:
            await pull.pull_now()
            return EmptyResponse()
        except Exception as exc:  # pylint: disable=broad-except
            _LOGGER.exception("While triggering the poll an error occurred.")
            raise HTTPException(
                status_code=500,
                detail="While triggering the poll an error occurred."
            ) from exc

    def attach(self, fastapi: FastAPI) -> None:
        fastapi.post(
            path="/trigger",
            response_class=JSONResponse,
            responses={
                500: {"description": "Internal Server Error"}
            }
        )(self.endpoint)



"""The actual application wrapper around tasks and engine."""
from typing import Optional

from typeguard import typechecked

from pnp.api import RestAPI
from pnp.api.endpoints import Trigger
from pnp.config import load_config, Configuration
from pnp.engines import DEFAULT_ENGINE, Engine
from pnp.models import TaskSet
from pnp.selector import PayloadSelector
from pnp.utils import Loggable, ReprMixin


class Application(Loggable, ReprMixin):
    """The wrapper that knows about tasks and engine."""

    __REPR_FIELDS__ = ['_api', '_engine', '_tasks']

    @typechecked
    def __init__(self, config: Configuration):
        self.config = config
        if not self.config.engine:
            self.config.engine = DEFAULT_ENGINE

        self._tasks = config.tasks
        self._engine = config.engine

        self._api = None  # type: Optional[RestAPI]
        if config.api:
            self._api = RestAPI()
            self._api.create_api(
                enable_metrics=config.api.enable_metrics,
            )
            Trigger(config.tasks).attach(self._api.fastapi)

    @property
    def api(self) -> Optional[RestAPI]:
        """Return the api that is configured. If no api is available `None` is returned."""
        return self._api

    @property
    def engine(self) -> Engine:
        """Return the engine that is used to schedule and run the tasks."""
        return self._engine

    @property
    def tasks(self) -> TaskSet:
        """Return the tasks that are configured for this application."""
        return self._tasks

    @classmethod
    def from_file(cls, file_path: str) -> 'Application':
        """
        Loads the application from a configuration file.

        Args:
            file_path (str): Where the configuration file is located.
        """
        config = load_config(str(file_path))
        PayloadSelector.instance.register_udfs(config.udfs)  # pylint: disable=no-member
        app = Application(config)
        return app



"""Simple pull plugins"""

import sys
from datetime import datetime

from pnp import validator
from pnp.config import load_pull_from_snippet
from pnp.plugins.pull import AsyncPull, AsyncPullNowMixin, SyncPolling
from pnp.typing import Payload
from pnp.utils import make_list, parse_duration_literal_float


class Count(AsyncPull):
    """
    Emits every `wait` seconds a counting value which runs from `from_cnt` to `to_cnt`.
    If `to_cnt` is None will to count to infinity (or more precise sys.maxsize).

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/simple.Count/index.md

    """
    __REPR_FIELDS__ = ['from_cnt', 'interval', 'to_cnt']

    DEFAULT_INTERVAL = 5

    def __init__(self, from_cnt=0, to_cnt=None, wait=None, interval=None, **kwargs):
        super().__init__(**kwargs)
        self.from_cnt = int(from_cnt)
        self.to_cnt = to_cnt and int(to_cnt)
        self.interval = float(parse_duration_literal_float(
            interval or wait or self.DEFAULT_INTERVAL
        ))

    @property
    def can_exit(self) -> bool:
        return True  # pragma: no cover

    async def _pull(self):
        to_cnt = sys.maxsize
        if self.to_cnt:
            to_cnt = self.to_cnt + 1
        for i in range(self.from_cnt, to_cnt):
            await self._sleep(self.interval)
            self.notify(i)
            if self.stopped:
                break


class Cron(SyncPolling):
    """
    Cron-like triggering of dependent pushes.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/simple.Cron/index.md
    """

    __REPR_FIELDS__ = 'expressions'

    def __init__(self, expressions, **kwargs):
        super().__init__(interval=60, instant_run=False, **kwargs)
        self.expressions = make_list(expressions)

        from cronex import CronExpression
        self.jobs = [CronExpression(expression) for expression in self.expressions]

    def _configure_scheduler(self, scheduler, callback):
        scheduler.every().minute.at(":00").do(callback)

    def _poll(self):
        for job in self.jobs:
            dtime = datetime.now()
            if job.check_trigger((dtime.year, dtime.month, dtime.day, dtime.hour, dtime.minute)):
                self.notify(job.comment)

    async def _pull_now(self) -> None:
        for job in self.jobs:
            self.notify(job.comment)


class CustomPolling(SyncPolling):
    """
    Calls the specified callable every `interval`. The result of the callable is simply returned.
    This plugin is basically for _internal_ use only.
    """
    def __init__(self, scheduled_callable, **kwargs):
        super().__init__(**kwargs)
        self.scheduled_callable = scheduled_callable
        validator.is_function(scheduled_callable=self.scheduled_callable)

    def _poll(self):
        return self.scheduled_callable()


class RunOnce(AsyncPull):
    """
    Takes another valid `plugins.pull.Polling` component and immediately executes it and ventures
    down the given `plugins.push` components. If no component is given it will simple execute the
    push chain.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/simple.RunOnce/index.md
    """

    __REPR_FIELDS__ = 'model'

    def __init__(self, poll=None, **kwargs):
        super().__init__(**kwargs)
        self.model = None
        self.wrapped = None
        if poll:
            self.model = load_pull_from_snippet(poll, base_path=self.base_path, name=self.name)
            self.wrapped = self.model.instance

            if not self.wrapped.supports_pull_now:
                raise TypeError(
                    "Pull does not support pull_now()."
                    " Implement PullNowMixin / AsyncPullNowMixin for support"
                )

    @property
    def can_exit(self) -> bool:
        return True  # pragma: no cover

    async def _pull(self) -> None:
        if not self.wrapped:
            self.notify({})  # Just notify about an empty dict
        else:
            def callback(plugin, payload: Payload):
                _ = plugin  # Fake usage
                self.notify(payload)

            self.wrapped.callback(callback)
            await self.wrapped.pull_now()


class Repeat(AsyncPull, AsyncPullNowMixin):
    """
    Emits every `wait` seconds the same `repeat`.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/simple.Repeat/index.md

    """
    __REPR_FIELDS__ = ['interval', 'repeat']

    DEFAULT_INTERVAL = 5

    def __init__(self, repeat, wait=None, interval=None, **kwargs):
        super().__init__(**kwargs)
        self.repeat = repeat
        self.interval = float(parse_duration_literal_float(
            interval or wait or self.DEFAULT_INTERVAL
        ))

    async def _pull(self):
        while not self.stopped:
            await self._sleep(self.interval)
            await self._pull_now()

    async def _pull_now(self):
        self.notify(self.repeat)



"""File system related plugins."""

import os
import time

from pnp import validator
from pnp.plugins import load_optional_module, PluginStoppedError
from pnp.plugins.pull import SyncPull, SyncPolling
from pnp.utils import make_list, load_file, FILE_MODES, Debounce


class FileSystemWatcher(SyncPull):
    """
    Watches the given directory for changes like created, moved, modified and deleted files.
    If ignore_directories is set to False, then directories will be reported as well.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/fs.FileSystemWatcher/index.md
    """
    __REPR_FIELDS__ = [
        'base64', 'case_sensitive', 'defer_modified', 'events', 'ignore_directories',
        'ignore_patterns', 'load_file', 'mode', 'path', 'patterns', 'recursive'
    ]

    EXTRA = 'fswatcher'
    EVENT_TYPE_MOVED = 'moved'
    EVENT_TYPE_DELETED = 'deleted'
    EVENT_TYPE_CREATED = 'created'
    EVENT_TYPE_MODIFIED = 'modified'
    EVENT_TYPES = [EVENT_TYPE_MOVED, EVENT_TYPE_DELETED, EVENT_TYPE_CREATED, EVENT_TYPE_MODIFIED]

    # pylint: disable=redefined-outer-name
    def __init__(
        self, path, recursive=True, patterns=None, ignore_patterns=None,
        ignore_directories=False, case_sensitive=False, events=None, load_file=False,
        mode='auto', base64=False, defer_modified=0.5, **kwargs
    ):
        super().__init__(**kwargs)
        self.path = path
        validator.is_directory(path=self.path)
        self.recursive = bool(recursive)
        self.patterns = make_list(patterns)
        self.ignore_patterns = make_list(ignore_patterns)
        self.ignore_directories = bool(ignore_directories)
        self.case_sensitive = bool(case_sensitive)
        self.load_file = bool(load_file)
        self.base64 = bool(base64)
        self.events = make_list(events)
        if self.events:
            validator.subset_of(self.EVENT_TYPES, events=self.events)
        self.mode = mode
        validator.one_of(FILE_MODES, mode=self.mode)
        self.defer_modified = float(defer_modified)
        if self.defer_modified < 0.0:
            self.defer_modified = 0.5

    def _pull(self):
        wdog = load_optional_module('watchdog.observers', self.EXTRA)
        wev = load_optional_module('watchdog.events', self.EXTRA)

        that = self

        class _EventHandler(wev.PatternMatchingEventHandler):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.dispatcher = {}

            @staticmethod
            def _read_file_from_event(event):
                if (event.event_type != that.EVENT_TYPE_DELETED
                        and not getattr(event, 'is_directory', False)):
                    file_name = (
                        getattr(event, 'dest_path', None)
                        or getattr(event, 'src_path', None)
                    )
                    if file_name is not None:
                        return load_file(file_name, that.mode, that.base64)
                return None

            def stop_dispatcher(self):
                """Stops any pending debounces when the plugin is going to stop."""
                # We cannot alter the dictionary during iteration - so we have to get all relevant
                # debounces beforehand...
                candidates = [debounce for _, debounce in self.dispatcher.items()]
                # ... and then loop it over outside the iterator context
                for debounce in candidates:
                    debounce.execute_now()

            def _deferred_notify(self, modified_file, payload):
                # Let's remove the dispatcher instance - it is done
                self.dispatcher.pop(modified_file, None)
                # trigger the actual notification event
                that.notify(payload)

            def _notify(self, event, payload):
                if event.event_type == that.EVENT_TYPE_MODIFIED:
                    # There might be multiple flushes of a file before it is completely written to
                    # disk. Each flush will raise a modified event...
                    # Let's wait for more modified events...
                    if that.defer_modified <= 0:
                        that.notify(payload)
                    else:
                        modified_file = payload['source']
                        defer_fun = self.dispatcher.get(modified_file, None)
                        if defer_fun is None:
                            defer_fun = Debounce(self._deferred_notify, that.defer_modified)
                            self.dispatcher[modified_file] = defer_fun
                        defer_fun(modified_file, payload)
                        that.logger.debug(
                            "Event of modified_file '%s' is deferred for %s",
                            modified_file, that.defer_modified
                        )
                else:
                    defer_fun = self.dispatcher.get(payload['source'], None)
                    # There might be some deferred modified event... Lets check and send it to keep
                    # the correct sequence. Assuming the file is closed and finished by now...
                    if defer_fun is not None:
                        defer_fun.execute_now()
                    that.notify(payload)

            def on_any_event(self, event):
                """Callback for watchdog."""
                if that.events is None or event.event_type in that.events:
                    payload = {
                        'operation': event.event_type,
                        'is_directory': getattr(event, 'is_directory', False),
                        'source': getattr(event, 'src_path', None),
                        'destination': getattr(event, 'dest_path', None)
                    }
                    that.logger.info("Got '%s'", payload)
                    if that.load_file:
                        file_envelope = self._read_file_from_event(event)
                        if file_envelope is not None:
                            payload['file'] = file_envelope

                    self._notify(event, payload)

        observer = wdog.Observer()
        handler = _EventHandler(
            patterns=self.patterns,
            ignore_patterns=self.ignore_patterns,
            ignore_directories=self.ignore_directories,
            case_sensitive=self.case_sensitive
        )
        observer.schedule(handler, self.path, recursive=self.recursive)
        observer.start()

        while not self.stopped:
            time.sleep(1)

        observer.stop()
        handler.stop_dispatcher()
        observer.join()


class Size(SyncPolling):
    """
    Periodically determines the size of the specified files or directories in bytes.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/fs.FileSize/index.md
    """

    __REPR_FIELDS__ = ['fail_on_error', 'file_paths']

    def __init__(self, paths, fail_on_error=True, **kwargs):
        super().__init__(**kwargs)
        if isinstance(paths, dict):
            self.file_paths = paths  # alias -> file path mapping
        else:
            # No explicit alias: Use the basename of the file
            self.file_paths = {os.path.basename(str(fp)): str(fp) for fp in make_list(paths)}
        self.fail_on_error = bool(fail_on_error)

    @staticmethod
    def _file_size(path):
        return os.path.getsize(path)

    def _dir_size(self, path):
        total_size = 0
        for dirpath, _, filenames in os.walk(path):
            for fname in filenames:
                fp = os.path.join(dirpath, fname)
                try:
                    total_size += self._file_size(os.path.realpath(fp))
                except FileNotFoundError:
                    self.logger.warning(
                        "File %s does not exists when calculating the directory size", fp)
                finally:
                    if self.stopped:
                        raise PluginStoppedError(
                            "Aborted calculation of directory size, due to stop of plugin")

        return total_size

    def _size(self, path):
        try:
            if os.path.isdir(path):
                return self._dir_size(path)
            return self._file_size(path)
        except (FileNotFoundError, NotADirectoryError):
            if self.fail_on_error:
                raise
            return None

    def _poll(self):
        return {
            alias: self._size(fp)
            for alias, fp in self.file_paths.items()
        }



"""Base stuff for pushes."""

import asyncio
import copy
import functools
import inspect
from abc import abstractmethod
from typing import Any, Callable, Optional, Dict, Iterable, Union, cast, List, Tuple, Awaitable

from pnp import utils
from pnp import validator
from pnp.plugins import Plugin
from pnp.shared.async_ import run_sync
from pnp.typing import Envelope, Payload

# Push function typing alias
PushFunction = Union[Callable[..., Payload], Callable[..., Awaitable[Payload]]]


class PushExecutionError(Exception):
    """Is raised by push-plugins when the execution flow has failed."""


def enveloped(
    fun: PushFunction
) -> Callable[['Push', Payload], Payload]:
    """Decorator to split the envelope and the actual payload. This is an but a convenience
    decorator for `envelope_payload` of the `PushBase` class."""

    validator.is_function(fun=fun)

    def _call(self: 'Push', payload: Payload) -> Payload:
        validator.is_instance(Push, self=self)
        envelope, real_payload = self.envelope_payload(payload)
        return fun(
            self,
            envelope=envelope,
            payload=real_payload
        )

    if asyncio.iscoroutinefunction(fun):
        @functools.wraps(fun)
        async def _wrapper(self: 'Push', payload: Payload) -> Payload:
            return await _call(self, payload)
        return _wrapper

    return functools.wraps(fun)(_call)


def parse_envelope(value: str) -> PushFunction:
    """Decorator the parse the given value-key from the envelope. This is but a convenience
    decorator / wrapper for `_parse_envelope_value` of the `PushBase` class."""

    validator.is_instance(str, value=value)

    def _inner(fun: PushFunction) -> PushFunction:
        validator.is_function(fun=fun)

        def _call(
                self: 'Push', envelope: Optional[Envelope] = None, payload: Payload = None,
                **kwargs: Any
        ) -> Payload:
            validator.is_instance(Push, self=self)
            parsed = self._parse_envelope_value(value, envelope=envelope)

            new_kwargs = {
                value: parsed,
                'payload': payload
            }
            if envelope is not None:
                new_kwargs['envelope'] = envelope

            return fun(self, **{**new_kwargs, **kwargs})

        if asyncio.iscoroutinefunction(fun):
            @functools.wraps(fun)
            async def _wrapper(
                    self: 'Push', envelope: Optional[Envelope] = None, payload: Payload = None,
                    **kwargs: Any
            ) -> Payload:
                return await _call(self, envelope, payload, **kwargs)
            return _wrapper

        return functools.wraps(fun)(_call)
    return _inner


def drop_envelope(fun: PushFunction) -> PushFunction:
    """Decorator to drop the envelope from the arguments before calling the actual `push` method to
    make the linter happy again if necessary (unused-argument)."""

    def _call(self: 'Push', *args: Any, **kwargs: Any) -> Any:
        validator.is_instance(Push, self=self)
        kwargs.pop('envelope', None)
        return fun(self, *args, **kwargs)

    if asyncio.iscoroutinefunction(fun):
        @functools.wraps(fun)
        async def _wrapper(self: 'Push', *args: Any, **kwargs: Any) -> Any:
            return await _call(self, *args, **kwargs)
        return _wrapper

    return functools.wraps(fun)(_call)


def _lookup(instance: object, lookups: Union[str, Iterable[str]]) -> Optional[Any]:
    """
    Will lookup the instance for the given attribute names in `lookups`.
    Returns the first occurrence found.

    Examples:

        >>> class Foo:
        ...     def __init__(self):
        ...         self.a = 1
        ...         self.b = 2
        ...         self.c = 3
        ...         self.__a = 99
        >>> instance = Foo()
        >>> _lookup(instance, 'a')
        1
        >>> _lookup(instance, '__a')
        99
        >>> _lookup(instance, ['c', 'b'])
        3
        >>> _lookup(instance, ['d', 'b'])
        2
        >>> print(_lookup(instance, ['x', 'y', 'z']))
        None

    Args:
        instance: The class to lookup for the given attribute names.
        lookups: Attribute names to lookup.

    Returns:
        Returns the value of the first found attribute name in lookups. If none is found, simply
        None is returned.
    """
    lookups = cast(List[str], utils.make_list(lookups) or [])
    # There might be some name mangling for private attributes - add them as lookups
    privates = ['_{classname}{attrname}'.format(classname=type(instance).__name__, attrname=x)
                for x in lookups if x.startswith('__') and not x.endswith('__')]
    lookups = lookups + privates
    _nothing = object()
    for lookup in lookups:
        val = getattr(instance, lookup, _nothing)
        if val is not _nothing:
            return val
    return None


class Push(Plugin):
    """Base class for all push plugins."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._assert_push_compat()

    def _assert_push_compat(self) -> None:
        self._assert_abstract_compat((SyncPush, AsyncPush))
        self._assert_fun_compat('_push')

    def _parse_envelope_value(
            self, name: str, envelope: Optional[Envelope] = None,
            parse_fun: Optional[Callable[[Any], Any]] = None,
            instance_lookup_fun: Optional[Callable[..., Optional[Any]]] = None
    ) -> Any:
        """
        Parse the envelope for the given `name`. If present in the envelope the extracted value will
        be tested / parsed by the given `parse_fun`. If no `parse_fun` is explicitly given, it will
        be determined if this instance provides a `parse_<name>`, `_parse_<name>` or a
        `__parse_<name>`. If yes the function is called; otherwise the value is returned as is.

        If the value is not present in the envelope the current instance will be probed for
        `<name>`, `_<name>` or `__<name>`. If an instance variable is present it will be returned
        unvalidated / unparsed (we assume that happened previously). If no variable is present
        simply `None` will be returned.

        Args:
            name: Name of the attribute to lookup in the envelope / instance.
            envelope: Envelope of the payload.
            parse_fun: Custom function to validate / parse. If not given it will be determined
                automagically.
            instance_lookup_fun: Custom function to perform instance variable lookups.
                If not given a reasonably default will be used.
        """
        validator.is_instance(str, name=name)
        validator.is_instance(dict, allow_none=True, envelope=envelope)
        if parse_fun:
            validator.is_function(parse_fun=parse_fun)
        if instance_lookup_fun:
            validator.is_function(instance_lookup_fun=instance_lookup_fun)

        lookups = cast(
            Dict[str, str], utils.make_public_protected_private_attr_lookup(name, as_dict=True)
        )

        if envelope is None or name not in envelope:
            return (
                instance_lookup_fun(name)
                if instance_lookup_fun is not None
                else _lookup(self, list(lookups.values()))  # pylint: disable=no-member
            )

        val = envelope[name]
        try:
            if parse_fun is None:
                public_attr_name = lookups['public']  # pylint: disable=invalid-sequence-index
                fun_base = 'parse_' + public_attr_name
                parse_fun_names = utils.make_public_protected_private_attr_lookup(fun_base)
                parse_fun = _lookup(self, parse_fun_names)
                if parse_fun is None:
                    return val
            return parse_fun(val)
        except (ValueError, TypeError):
            self.logger.exception(
                "Cannot parse value for '%s' from envelope",
                name
            )
            return (
                instance_lookup_fun(name)
                if instance_lookup_fun is not None
                else _lookup(self, list(lookups.values()))  # pylint: disable=no-member
            )

    @staticmethod
    def envelope_payload(payload: Payload) -> Tuple[Envelope, Payload]:
        """
        A payload might be enveloped (but actually it even may not). The actual payload will be
        available via data or payload inside the dictionary. All other keys are the envelope.

        Args:
            payload: Payload (which might be enveloped or not).

        Returns:
            A tuple consisting of envelope (might be None) and the real payload.

        Examples:

            >>> SyncPush.envelope_payload(1234)  # No envelope -> payload as is
            ({}, 1234)
            >>> (SyncPush.envelope_payload({'key1': 'val1', 'key2': 'val2'}) ==
            ...     ({}, {'key1': 'val1', 'key2': 'val2'}))
            True
            >>> SyncPush.envelope_payload({'data': 1234})  # Envelope with the data key only
            ({}, 1234)
            >>> SyncPush.envelope_payload({'payload': 1234})  # Works with 'payload' as well
            ({}, 1234)
            >>> SyncPush.envelope_payload({'envelope1': 'envelope', 'payload': 1234})
            ({'envelope1': 'envelope'}, 1234)
        """
        payload = copy.deepcopy(payload)

        if not isinstance(payload, dict):
            return {}, payload

        if 'data' in payload:
            real_payload = payload.pop('data')
        elif 'payload' in payload:
            real_payload = payload.pop('payload')
        else:
            return {}, payload  # There is no envelope

        # We popped the real payload -> all what is left in payload is the envelope
        return payload, real_payload

    async def push(self, payload: Payload) -> Payload:
        """Transforms the payload and/or pushes it to a sink."""
        push_fun = getattr(self, '_push')
        if inspect.iscoroutinefunction(push_fun):
            return await push_fun(payload)
        return await run_sync(push_fun, payload)


class SyncPush(Push):
    """Base class for all synchronous push plugins."""

    @abstractmethod
    def _push(self, payload: Payload) -> Payload:
        """
        This is where the hard work happens.

        Args:
            payload: The payload.
        """
        raise NotImplementedError()  # pragma: no cover


class AsyncPush(Push):
    """Base class for all asynchronous push plugins."""

    @abstractmethod
    async def _push(self, payload: Payload) -> Payload:
        """
        This is where the hard work happens.

        Args:
            payload: The payload.
        """
        raise NotImplementedError()  # pragma: no cover



"""Z-Way related pulls."""

import requests
from requests.auth import HTTPBasicAuth

from pnp.plugins.pull import PollingError, SyncPolling


class ZwayPoll(SyncPolling):
    """
    Pulls the specified json content from the z-way rest api.
    The content is specified by the url, e.g. http://<host>:8083/ZWaveAPI/Run/devices will pull
    all devices and serve the result as a json.

    See Also:
        https://pnp.readthedocs.io/en/stable/plugins/index.html#zway-zwaypoll
    """

    __REPR_FIELDS__ = ['url', 'user']

    def __init__(self, url, user, password, **kwargs):
        super().__init__(**kwargs)
        self.url = str(url)
        self.user = str(user)
        self.password = str(password)

    def _poll(self):
        self.logger.debug("Polling url '%s'", self.url)
        reply = requests.get(self.url, auth=HTTPBasicAuth(self.user, self.password))
        if reply.status_code != requests.codes.ok:  # pylint: disable=no-member
            raise PollingError(
                "Code {code}: {error}".format(code=reply.status_code, error=reply.text)
            )

        try:
            return reply.json()
        except ValueError:
            # No valid json, try text
            return reply.text



"""Base classes and interfaces for endpoints."""

from fastapi import FastAPI


class Endpoint:
    """Base class for endpoints."""

    def attach(self, fastapi: FastAPI) -> None:
        """Attach this endpoint(s) to the fastapi serving component."""
        raise NotImplementedError()



"""MQTT related utility classes."""

import json
import logging
from typing import Optional, Any

from pnp.typing import Payload
from pnp.utils import try_parse_int

_LOGGER = logging.getLogger(__name__)


class MQTTBase:
    """MQTT base class for publishing message to a mqtt broker."""

    def __init__(
        self, host: str, port: int = 1883, user: Optional[str] = None,
        password: Optional[str] = None, qos: int = 0, **kwargs: Any
    ):
        super().__init__(**kwargs)  # type: ignore
        self.host = str(host)
        self.port = int(port)
        self.user = user and str(user)
        self.password = password and str(password)
        self.qos = self._parse_qos(qos)

    @staticmethod
    def _parse_qos(val: int) -> int:
        pval = try_parse_int(val)
        if pval is None:
            _LOGGER.warning("QOS level of '%s' is not int parsable. Defaulting to 0", val)
            return 0
        if pval < 0:
            _LOGGER.warning("QOS level of '%s' is invalid. Defaulting to 0", val)
            return 0
        if pval > 2:
            _LOGGER.warning("QOS level of '%s' is invalid. Defaulting to 2", val)
            return 2
        return pval

    def _publish(
        self, payload: Payload, topic: str, retain: bool = False, qos: Optional[int] = None
    ) -> None:
        if isinstance(payload, (dict, list, tuple)):
            payload = json.dumps(payload)

        auth = None
        if self.user:
            auth = dict(username=self.user, password=self.password)

        if qos is None:
            qos = self.qos

        import paho.mqtt.publish as publish
        publish.single(
            topic=topic,
            payload=payload,
            hostname=self.host,
            port=self.port,
            retain=retain,
            auth=auth,
            qos=qos
        )
        _LOGGER.debug(
            "Published message on '%s' @ %s:%s with qos=%s. Payload='%s'",
            topic, self.host, self.port, qos, payload
        )



"""Config Loader for yaml files."""

import os
from functools import partial
from typing import Any, Dict, Iterator, cast, Union, List, Iterable, Optional

import schema as sc
import yaml
from box import Box
from dictmentor import DictMentor, ext
from yamlinclude import YamlIncludeConstructor

from pnp.config._base import Configuration, ConfigLoader
from pnp.engines import Engine as RealEngine, RetryHandler
from pnp.models import UDFModel, PullModel, PushModel, TaskModel, TaskSet, APIModel
from pnp.plugins import load_plugin
from pnp.plugins.pull import Pull
from pnp.plugins.push import Push
from pnp.plugins.udf import UserDefinedFunction
from pnp.utils import make_list

# Type alias that represents a yaml config snippet
PartialConfig = Any

# Type alias that represents yaml push config snippet
PushConfig = Dict[str, Any]


class Schemas:
    """Collection of validation schemas for different components."""

    plugin_args_name = "args"
    plugin_name = "plugin"

    # Single pull
    Pull = sc.Schema({
        plugin_name: sc.Use(str),
        sc.Optional(plugin_args_name, default={}): {
            sc.Optional(str): object
        },
    })

    # Single push block

    push_selector_name = "selector"
    push_unwrap_name = "unwrap"
    push_deps_name = "deps"

    Push = sc.Schema({
        plugin_name: sc.Use(str),
        sc.Optional(push_selector_name, default=None): sc.Or(object, None),
        sc.Optional(push_unwrap_name, default=False): bool,
        sc.Optional(plugin_args_name, default={}): {
            sc.Optional(str): object
        },
        # deps is a list of push_schemas as well, but cannot declare recursive schemas
        sc.Optional(push_deps_name, default=list()): sc.And(object, sc.Use(make_list))
    })

    # Multiple pushes
    PushList = sc.Schema([Push])

    # Single task (one pull and multiple related pushes)
    task_name = "name"
    task_pull_name = "pull"
    task_push_name = "push"

    Task = sc.Schema({
        task_name: sc.Use(str),
        task_pull_name: Pull,
        task_push_name: sc.And(sc.Or(PushList, Push), sc.Use(make_list))
    })

    # Allow configuration of a single task or a list of tasks
    TaskList = sc.Schema(
        sc.Or(
            sc.And(Task, sc.Use(make_list)),
            [Task]
        )
    )

    # Engine
    Engine = RealEngine

    # API
    api_port_name = "port"
    api_endpoint_name = "endpoints"
    api_endpoint_metrics = "metrics"

    api_endpoint_defaults = {
        api_endpoint_metrics: False
    }

    API = sc.Schema({
        api_port_name: sc.Use(int),
        sc.Optional(api_endpoint_name, default=api_endpoint_defaults): {
            sc.Optional(api_endpoint_metrics, False): sc.Use(bool)
        }
    })

    # User Defined Functions
    udf_name = "name"

    UDFS = sc.Schema([{
        udf_name: sc.Use(str),
        plugin_name: sc.Use(str),
        sc.Optional(plugin_args_name): sc.Or({
            str: object
        }, None)
    }])

    # Global settings

    global_engine_name = "engine"
    global_api_name = "api"
    global_udfs_name = "udfs"
    global_tasks_name = "tasks"

    GlobalSettings = sc.Schema({
        sc.Optional(sc.Or("anchors", "anchor", "ref", "refs", "alias", "aliases")): object,
        sc.Optional(global_engine_name, default=None): Engine,
        sc.Optional(global_api_name, default=None): API,
        sc.Optional(global_udfs_name, default=None): UDFS,
        global_tasks_name: TaskList
    })

    Root = sc.Schema(sc.And(
        sc.Or(
            # Tasks with additional settings and aliases (yaml only)
            GlobalSettings,
            # Tasks without settings or aliases - just a list of tasks (backwards compatibility)
            TaskList,
        ),
        # In case of list only we create a tasks node - so we know it's always there
        sc.Use(lambda x: x if 'tasks' in x else dict(tasks=x))
    ))


def _custom_yaml_constructor(loader: Any, node: Any, clstype: Any) -> Any:
    """YAML custom constructor. Necessary to create an engine and retry handler."""
    args = loader.construct_mapping(node, deep=True)
    if 'type' not in args:
        raise ValueError("You have to specify a 'type' when instantiating a engine with !engine")
    clazz_name = args.pop('type')
    return load_plugin(clazz_name, clstype, **args)


def _custom_env_tag(loader: Any, node: Any) -> Any:
    if not isinstance(node, yaml.nodes.ScalarNode):
        raise yaml.constructor.ConstructorError(
            None, None,
            f'expected a scalar or sequence node, but found {node.id}',
            node.start_mark
        )

    node_value = loader.construct_scalar(node)
    sep_index = node_value.find(':=')
    if sep_index >= 1:
        envvar = node_value[:sep_index]
        default = node_value[sep_index + 2:]
    else:
        envvar = node_value
        default = None

    value = os.environ.get(envvar, default)
    if value is None:
        raise EnvironmentError(f"Environment variable '{envvar}' not found and no default set")

    tag = loader.resolve(yaml.nodes.ScalarNode, value, (True, False))
    return loader.construct_object(yaml.nodes.ScalarNode(tag, value))


def _mk_pull(task_config: Box, **extra: Any) -> PullModel:
    """Make a pull out of a task configuration."""
    name = '{}_pull'.format(task_config[Schemas.task_name])
    pull_args = task_config[Schemas.task_pull_name][Schemas.plugin_args_name]
    plugin = task_config[Schemas.task_pull_name][Schemas.plugin_name]
    args = {'name': name, **extra, **pull_args}
    return PullModel(instance=cast(Pull, load_plugin(
        plugin_path=plugin,
        plugin_type=Pull,
        instantiate=True,
        **args
    )))


def _mk_push(task_config: Box, **extra: Any) -> List[PushModel]:
    """Make one or more pushes out of task configuration."""
    def _many(pushlist: List[Box], prefix: str) -> Iterable[PushModel]:
        for i, push in enumerate(pushlist):
            push_name = '{prefix}_{i}'.format(
                i=i,
                prefix=prefix
            )
            args = {'name': push_name, **extra, **push[Schemas.plugin_args_name]}
            unwrap = getattr(push, Schemas.push_unwrap_name, False)
            yield PushModel(
                instance=cast(Push, load_plugin(
                    plugin_path=push[Schemas.plugin_name],
                    plugin_type=Push,
                    instantiate=True,
                    **args
                )),
                selector=push[Schemas.push_selector_name],
                unwrap=unwrap,
                deps=list(_many(push[Schemas.push_deps_name], push_name))
            )
    pushes = task_config[Schemas.task_push_name]
    prefix = task_config[Schemas.task_name] + "_push"
    return list(_many(pushes, prefix))


def _mk_udf(udf_config: Box) -> UDFModel:
    if not isinstance(udf_config, Box):
        udf_config = Box(udf_config)
    udf_type = cast(
        Union[str, type],
        'callable' if not hasattr(udf_config, Schemas.plugin_args_name) else UserDefinedFunction
    )
    instantiate = hasattr(udf_config, Schemas.plugin_args_name)
    kwargs = udf_config.get(Schemas.plugin_args_name) or {}
    fun = load_plugin(
        plugin_path=udf_config[Schemas.plugin_name],
        plugin_type=udf_type,
        instantiate=instantiate,
        **{'name': udf_config[Schemas.udf_name], **kwargs}
    )
    return UDFModel(name=udf_config[Schemas.udf_name], callable=cast(UserDefinedFunction, fun))


class YamlConfigLoader(ConfigLoader):
    """Configuration loader from a yaml file."""

    def _validate_push_deps(self, push: PushConfig) -> Iterator[PushConfig]:
        """Validates a push against the schema. This is done recursively."""
        _ = self  # Fake usage
        for dependency in push[Schemas.push_deps_name]:
            validated_push = Schemas.Push.validate(dependency)
            validated_push[Schemas.push_deps_name] = list(self._validate_push_deps(validated_push))
            yield validated_push

    def _add_constructors(self, base_path: str) -> None:
        _ = self  # Fake usage
        yaml.SafeLoader.add_constructor(  # type: ignore
            "!engine", partial(_custom_yaml_constructor, clstype=RealEngine)
        )
        yaml.SafeLoader.add_constructor(  # type: ignore
            "!retry", partial(_custom_yaml_constructor, clstype=RetryHandler)
        )
        yaml.SafeLoader.add_constructor(  # type: ignore
            "!include", YamlIncludeConstructor(base_dir=base_path)
        )
        yaml.SafeLoader.add_constructor(  # type: ignore
            "!env", _custom_env_tag
        )

    def _augment(self, configuration: PartialConfig, base_path: str) -> Any:
        """Augments the configuration by using dictmentor with `Environment`,
        `ExternalResource` and `ExternalYamlResource` plguins."""
        _ = self  # Fake usage
        assert isinstance(base_path, str)

        # pnp configuration is probably of list of tasks. dictmentor needs a dict ...
        # ... let's fake it ;-)
        cfg = dict(fake_root=configuration)

        mentor = DictMentor(
            ext.Environment(fail_on_unset=True),
            ext.ExternalResource(base_path=base_path),
            ext.ExternalYamlResource(base_path=base_path)
        )

        # Remove the faked dictionary as root level
        return mentor.augment(cfg)['fake_root']

    def _api_from_config(self, config: Box) -> Optional[APIModel]:
        _ = self  # Fake usage

        api = config.get(Schemas.global_api_name)
        if not api:
            return None

        return APIModel(
            port=api[Schemas.api_port_name],
            enable_metrics=api[Schemas.api_endpoint_name].get(Schemas.api_endpoint_metrics, False)
        )

    def _tasks_from_config(self, config: Box, base_path: Optional[str] = None) -> TaskSet:
        """Create a task from a task configuration."""
        _ = self  # Fake usage

        extra_kwargs = {'base_path': base_path}
        res = {}
        for task in config[Schemas.global_tasks_name]:
            instance = TaskModel(
                name=task[Schemas.task_name],
                pull=_mk_pull(task, **extra_kwargs),
                pushes=list(_mk_push(task, **extra_kwargs))
            )
            res[instance.name] = instance

        return res

    def _udfs_from_config(self, config: Box) -> List[UDFModel]:
        """Creates a UDFModel from a udf configuration snippet."""
        _ = self  # Fake usage

        udfs = config.get(Schemas.global_udfs_name)
        if not udfs:
            return []

        return [_mk_udf(udf_config) for udf_config in udfs or []]

    @classmethod
    def supported_extensions(cls) -> Iterable[str]:
        return ['json', 'yaml', 'yml']

    def load_pull_from_snippet(self, snippet: Any, name: str, **extra: Any) -> PullModel:
        pull_config = Schemas.Pull.validate(snippet)
        return _mk_pull(Box({'name': name, 'pull': pull_config}), **extra)

    def load_config(self, config_file: str) -> Configuration:
        config_file = str(config_file)
        base_path = os.path.abspath(os.path.dirname(config_file))

        with open(config_file, 'r') as fp:
            # Custom yaml constructors: !engine and !retry
            self._add_constructors(base_path)
            cfg = yaml.safe_load(fp)

        validated = Schemas.Root.validate(self._augment(cfg, base_path))
        # We need to validate each push again against the push-schema.
        # Cause the dependencies of push are pushes as well and we cannot define
        # a recursive schema in one go.
        for pull in validated[Schemas.global_tasks_name]:
            for push in pull[Schemas.task_push_name]:
                push[Schemas.push_deps_name] = list(self._validate_push_deps(push))
        config = Box(validated)

        return Configuration(
            api=self._api_from_config(config),
            engine=config.get(Schemas.global_engine_name) or None,
            udfs=self._udfs_from_config(config),
            tasks=self._tasks_from_config(config, base_path)
        )



"""Some typing stuff."""

from typing import Any, Dict, Union, List, Callable, Optional, TypeVar

# Generic Type T
T = TypeVar('T')

# Callable that allows any arguments and returns some value
AnyCallable = Callable[..., Any]

# Duration literal can be either something like 1h, 30 (for seconds) or 1.5 (seconds, but float)
DurationLiteral = Union[str, int, float]

# Metadata for the pull payload
Envelope = Dict[str, Any]

# Pull payload. Just for make things more readable
Payload = Any

# A valid expression for a selector is either
# * direct python code (str)
# * A dictionary with lambdas or literals (Dict[str, str])
# * A list of lambdas or literals
SelectorExpression = Optional[Union[str, Dict[Any, Any], List[Any]]]



"""Contains base classes for engines."""

import copy
from abc import abstractmethod, ABCMeta
from datetime import datetime
from typing import Any, Callable, Optional, Coroutine, Iterable

import pydantic
import typeguard
from typeguard import check_argument_types

from pnp import validator
from pnp.models import TaskSet, PushModel
from pnp.selector import PayloadSelector
from pnp.shared.async_ import run_sync
from pnp.typing import Payload
from pnp.utils import (
    Loggable,
    Singleton,
    parse_duration_literal,
    DurationLiteral,
    is_iterable_but_no_str,
    ReprMixin
)


class NotSupportedError(Exception):
    """Is raised when a task is not supported by an engine."""


# Signature of push result callback
PushResultCallback = Callable[[Payload, PushModel], None]

# Signature of a on engine started callback
OnStartedCallback = Callable[[], Coroutine[Any, Any, None]]

# Signature of a on engine stopped callback
OnStoppedCallback = Callable[[], Coroutine[Any, Any, None]]


class Engine(Loggable, ReprMixin, metaclass=ABCMeta):
    """
    An engine is responsible for launching a defined set of tasks and the communication between
    pulls and and their associated pushes.

    A call to to an engine's `run(...)` method will block the calling thread until the engine
    decides the job is done (normally an external SIGTERM occurs)
    """
    __REPR_FIELDS__ = 'is_running'

    def __init__(self) -> None:
        self._is_running = False
        self._tasks: Optional[TaskSet] = None
        self._on_started_callback: Optional[Any] = None
        self._on_stopped_callback: Optional[Any] = None

    @property
    def is_running(self) -> bool:
        """Returns True if this engine is currently running; otherwise False."""
        return self._is_running

    @property
    def tasks(self) -> Optional[TaskSet]:
        """Returns the tasks that are currently running; otherwise (non running)
        None is returned."""
        return self._tasks

    @property
    def on_started_callback(self) -> Optional[OnStartedCallback]:
        """Returns the callback that is called when the engine has started."""
        return self._on_started_callback

    @on_started_callback.setter
    def on_started_callback(self, value: Optional[OnStartedCallback]) -> None:
        """Sets the callback that is called when the engine has started."""
        check_argument_types()
        self._on_started_callback = value

    @property
    def on_stopped_callback(self) -> Optional[OnStoppedCallback]:
        """Returns the callback that is called when the engine has stopped."""
        return self._on_stopped_callback

    @on_stopped_callback.setter
    def on_stopped_callback(self, value: Optional[OnStoppedCallback]) -> None:
        """Sets the callback that is called when engine has stopped."""
        check_argument_types()
        self._on_stopped_callback = value

    async def start(self, tasks: TaskSet) -> None:
        """Run the given task set inside the engine."""
        if self.is_running:
            return

        self._tasks = tasks
        await self._start(tasks)
        self._is_running = True

        if self._on_started_callback is not None:
            assert callable(self._on_started_callback)
            await self._on_started_callback()

    @abstractmethod
    async def _start(self, tasks: TaskSet) -> None:
        """Run the given task set inside the engine. Override in child classes to do the hard
        work."""
        raise NotImplementedError()

    async def stop(self) -> None:
        """Stop the engine."""
        if not self.is_running:
            return

        if not self.tasks:
            return

        await self._stop()
        self._tasks = None
        self._is_running = False

        if self._on_stopped_callback is not None:
            assert callable(self._on_stopped_callback)
            await self._on_stopped_callback()

    @abstractmethod
    async def _stop(self) -> None:
        """Stop the engine. Override in child classes."""
        raise NotImplementedError()


class RetryDirective(pydantic.BaseModel):
    """Contains directive information for engines on how to proceed in erroneous cases."""

    # If set to True, instructs the engine to abort the pull; otherwise retry the pull
    abort: bool = True
    # Instructs the engine to wait some time before retrying the pull again
    wait_for: int = 0
    # Just contextual information how many retries occurred so far
    retry_cnt: int = 0

    @pydantic.validator('wait_for', pre=True)
    @classmethod
    def _wait_for_duration_literal_converter(cls, value: Any) -> int:
        typeguard.check_type('wait_for', value, DurationLiteral)  # type: ignore
        return parse_duration_literal(value)


class RetryHandler(ReprMixin, Loggable, metaclass=ABCMeta):
    """
    A retry handler comes into play when a pull component exits unexpectedly whether by breaking
    the actual loop or by error. The `RetryHandler` decides how to proceed further (wait and retry
    the pull or more sophisticated logic).
    """
    __REPR_FIELDS__: Iterable[str] = []

    def __init__(self, **kwargs: Any):
        super().__init__()

        for unused_param in kwargs:
            self.logger.warning("Argument '%s' is unused", unused_param)

    @abstractmethod
    async def handle_error(self) -> RetryDirective:
        """
        Asks the retry handler to handle an error.

        Returns:
            An instance of `RetryDirective`.
        """
        raise NotImplementedError()


class NoRetryHandler(RetryHandler):
    """Will instruct the engine to abort and not to retry the pull."""

    async def handle_error(self) -> RetryDirective:
        return RetryDirective(abort=True, wait_for=0, retry_cnt=1)


class SimpleRetryHandler(RetryHandler):
    """Simply instructs the engine to wait for the given amount of time after an error."""

    __REPR_FIELDS__ = ['retry_wait', 'retry_count']

    def __init__(self, retry_wait: DurationLiteral = 60, **kwargs: Any):
        super().__init__(**kwargs)
        self.retry_wait = parse_duration_literal(retry_wait)
        self.retry_count = 0

    async def handle_error(self) -> RetryDirective:
        self.retry_count += 1
        return RetryDirective(abort=False, wait_for=self.retry_wait, retry_cnt=self.retry_count)


class LimitedRetryHandler(SimpleRetryHandler):
    """Instructs the engine to wait for the given amout of time after an error. If the specified
     `max_retries` is hit the engine is instructed to abort the `pull`."""

    __REPR_FIELDS__ = ['max_retries']

    def __init__(self, max_retries: Optional[int] = 3, **kwargs: Any):
        super().__init__(**kwargs)
        self.max_retries = max_retries and int(max_retries)  # type: Optional[int]

    def _eval_abort(self, retry_count: int) -> bool:
        if self.max_retries is None or self.max_retries < 0:
            return False
        return retry_count > self.max_retries

    async def handle_error(self) -> RetryDirective:
        await super().handle_error()
        abort = self._eval_abort(self.retry_count)
        return RetryDirective(abort=abort, wait_for=self.retry_wait, retry_cnt=self.retry_count)


class AdvancedRetryHandler(LimitedRetryHandler):
    """Works like the `LimitedRetryHandler` but will reset the retry count when a given amount of
    time between the current failure and the previous failure has passed."""

    __REPR_FIELDS__ = ['reset_retry_threshold']

    def __init__(self, reset_retry_threshold: DurationLiteral = 60, **kwargs: Any):
        super().__init__(**kwargs)
        # Reset retry_count after x seconds of successful running
        self.reset_retry_threshold = parse_duration_literal(reset_retry_threshold)
        self.last_error = None  # type: Optional[datetime]

    async def handle_error(self) -> RetryDirective:
        # Handles two cases:
        # 1. Initial value -> no retries so far, the next is 1
        # 2. Reset retry count because threshold has reached, next try is 1
        if (self.last_error is None
                or (datetime.now() - self.last_error).total_seconds() > self.reset_retry_threshold):
            self.retry_count = 0

        self.last_error = datetime.now()
        directive = await super().handle_error()

        return RetryDirective(
            abort=directive.abort,
            wait_for=self.retry_wait,
            retry_cnt=self.retry_count
        )


class PushExecutor(Loggable, Singleton):
    """
    Given a payload and the push this helper actually "executes" the push by passing the payload
    after the (optional) selector did some magic to it.
    """

    async def _internal(
            self, ident: str, payload: Payload, push: PushModel,
            result_callback: Optional[PushResultCallback] = None
    ) -> None:
        self.logger.debug("[%s] Selector: Applying '%s' to '%s'", ident, push.selector, payload)
        # The selector expression has no async support
        payload = await run_sync(
            PayloadSelector().eval_selector, push.selector, copy.deepcopy(payload)
        )

        if PayloadSelector().should_suppress(payload):
            self.logger.debug(
                "[%s] Selector evaluated to suppress literal. Skipping the push", ident
            )
            return

        self.logger.debug("[%s] Emitting '%s' to push '%s'", ident, payload, push.instance)
        push_result = await push.instance.push(payload)

        # Trigger any dependent pushes
        for dependency in push.deps:
            if result_callback:
                # Delegate work back to the engine
                self.logger.debug(
                    "[%s] Dependency callback is given. Delegating work back to engine", ident
                )
                result_callback(push_result, dependency)
            else:
                # No delegation callback defined: The executor will recursively call itself
                self.logger.debug(
                    "[%s] No callback is given. Recursively process dependencies", ident
                )
                await self.execute(ident, push_result, dependency)

    async def execute(
            self, ident: str, payload: Payload, push: PushModel,
            result_callback: Optional[PushResultCallback] = None
    ) -> None:
        """
        Executes the given push (in an asynchronous context) by passing the specified payload.
        In concurrent environments there might be multiple executions in parallel.
        You may specify an `id` argument to identify related execution steps in the logs.
        Use the `result_callback` when the engine can take care of dependent pushes as well.
        The result and a dependent push will be passed via the callback. If the callback is not
        specified the PushExecute will execute them in a recursive manner.

        Args:
            ident (str): ID to identify related execution steps in the logs
                (makes sense in concurrent environments).
            payload (Any): The payload to pass to the push.
            push (PushModel): The push instance that has to process the payload.
            result_callback (callable): See explanation above.
        """
        validator.is_instance(PushModel, push=push)

        if result_callback and not callable(result_callback):
            self.logger.warning(
                "Result callback is given, but is not a callable. Callback will be ignored."
            )
            result_callback = None

        if push.unwrap and is_iterable_but_no_str(payload):
            # Payload unwrapping
            length = len(payload)
            self.logger.debug("[%s] Unwrapping payload to %s individual items", ident, str(length))
            for item in payload:
                await self._internal(ident, item, push, result_callback)
        else:
            # Standard way
            await self._internal(ident, payload, push, result_callback)



"""Home assistant related push plugins"""

from pnp.plugins.push import SyncPush
from pnp.shared.hass import HassApi


class Service(SyncPush):
    """
    Calls a home assistant service providing the payload as service-data.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/hass.Service/index.md
    """

    __REPR_FIELDS__ = ['domain', 'service', 'timeout', 'url']

    def __init__(self, url, token, domain, service, timeout=10, **kwargs):
        super().__init__(**kwargs)
        self.url = str(url)
        self.token = str(token)
        self.domain = str(domain)
        self.service = str(service)
        self.timeout = timeout and int(timeout)
        if self.timeout <= 0:
            self.timeout = None  # Basically means no timeout
        self._client = HassApi(self.url, self.token, self.timeout)

    def _call(self, data):
        endpoint = 'services/{domain}/{service}'.format(
            domain=self.domain,
            service=self.service
        )

        try:
            return self._client.call(endpoint, method='post', data=data)
        except RuntimeError as exc:
            raise RuntimeError(
                "Failed to call the service {endpoint} @ {url}".format(
                    endpoint=endpoint,
                    url=self.url
                )
            ) from exc

    def _push(self, payload):
        self._call(payload)
        return payload



"""Global pnp api toolkit."""

import logging
from typing import Optional

from fastapi import FastAPI
from uvicorn.main import Server

from pnp import __version__
from pnp.api.endpoints import Ping, Health, SetLogLevel, PrometheusExporter, Version
from pnp.utils import Singleton

_LOGGER = logging.getLogger(__name__)


class APIError(RuntimeError):
    """Raised when some api error is experienced."""

    DEFAULT_MESSAGE = "Internal API error."

    def __init__(self, message: Optional[str] = None):
        super().__init__(message or self.DEFAULT_MESSAGE)


class APINotConfiguredError(APIError):
    """Raised when the api is not configured / enabled, but is necessary to perform a
    task."""

    DEFAULT_MESSAGE = "API is not configured, but is required to perform this task."


class APINotInitialized(APIError):
    """Raised when the api not initialized properly."""

    DEFAULT_MESSAGE = "API is not initialized properly. Try to `RestAPI.create_api(...)` first."


class RestAPI(Singleton):
    """API singleton."""

    def __init__(self) -> None:  # pylint: disable=super-init-not-called
        self.fastapi: Optional[FastAPI] = None
        self.port: Optional[int] = None
        self._fastapi_server: Optional[Server] = None

    def _assert_api(self) -> FastAPI:
        if not self.fastapi:
            raise APINotInitialized()
        return self.fastapi

    @property
    def enabled(self) -> bool:
        """Return True if the api is enabled; otherwise False."""
        return self.fastapi is not None

    def create_api(
        self, app_name: str = "pnp", enable_metrics: bool = True
    ) -> None:
        """
        Creates a fastAPI application to serve api requests.
        """
        self.fastapi = FastAPI(title=app_name, version=__version__)

        Ping().attach(self.fastapi)
        Health().attach(self.fastapi)
        SetLogLevel().attach(self.fastapi)
        Version().attach(self.fastapi)

        if bool(enable_metrics):
            PrometheusExporter(app_name).attach(self.fastapi)



"""Pull 'n' Push main entrypoint."""

import logging
import logging.config
import os
from functools import partial

import click
import coloredlogs
from ruamel import yaml
from sty import fg, bg, ef, rs

from pnp import __version__
from pnp.app import Application
from pnp.logo import PNP
from pnp.runner import Runner
from pnp.utils import get_first_existing_file


# Default log file name
DEFAULT_LOGGING_FILE_NAME = 'logging.yaml'

# Double space
DSPACE = " " * 2

# Print + flush
printf = partial(print, flush=True)


def _setup_logging(*candidates, log_level_override=None, env_key='PNP_LOG_CONF'):
    """Setup logging configuration"""
    log_file_path = get_first_existing_file(*candidates)
    env_path = os.getenv(env_key, None)
    if env_path:
        log_file_path = env_path
    if log_file_path and os.path.exists(log_file_path):
        log_file_path = os.path.abspath(log_file_path)
        with open(log_file_path, 'rt') as fhandle:
            config = yaml.safe_load(fhandle.read())
        logging.config.dictConfig(config)
        logging_conf = log_file_path
    else:
        if not log_level_override:
            log_level_override = "INFO"
        logging_conf = "Basic logging"

    coloredlogs.install(
        fmt='%(asctime)s %(name)s %(levelname)s %(message)s',
        level=log_level_override or logging.getLogger().level
    )
    return logging_conf


def _print_api_config(config):
    def helper():
        if not config.api:
            return f'{fg.yellow}No API configured{fg.rs}'
        return f'{fg.green}{config.api}{fg.rs}'

    printf(f"{ef.bold}API{rs.all}\n{DSPACE}{helper()}")


def _print_udf_config(config):
    print_str = f"{ef.bold}UDFs{rs.all}\n"
    if not config.udfs:
        print_str += f"{DSPACE}{fg.yellow}No UDFs configured{fg.rs}"
    else:
        all_udfs = [f"{DSPACE}- {fg.green}{str(udf)}{fg.rs}" for udf in config.udfs]
        print_str += "\n".join(all_udfs)
    printf(print_str)


def _print_engine_config(config):
    engine_str = f"{fg.green}{str(config.engine)}{fg.rs}"
    printf(f"{ef.bold}Engine{rs.all}\n{DSPACE}{engine_str}")


def _stringify_single_push(push, level):
    res = f"{DSPACE * level}- {fg.green}{push.instance}{fg.rs}"
    level += 1
    if push.selector:
        res += f"\n{DSPACE * level}{ef.bold}Selector{rs.all}: {push.selector}"
    if push.deps:
        res += "\n" + _stringify_push_config(push.deps, level, "Dependencies")
    return res


def _stringify_push_config(pushes, level=2, label="Push"):
    res = f"{DSPACE * level}{ef.bold}{label}{rs.all}\n"
    all_pushes = [_stringify_single_push(push, level + 1) for push in pushes]
    res += "\n".join(all_pushes)
    return res


def _print_tasks_config(config):
    print_str = f"{ef.bold}Tasks{rs.all}\n"
    all_tasks = []
    for task in config.tasks.values():
        task_str = ""
        task_str += f"{DSPACE}- {ef.bold}{task.name}{rs.all}\n"
        task_str += f"{DSPACE * 2}{ef.bold}Pull{rs.all}\n"
        task_str += f"{DSPACE * 3}{fg.green}{task.pull.instance}{fg.rs}\n"
        task_str += _stringify_push_config(task.pushes)
        all_tasks.append(task_str)
    print_str += "\n".join(all_tasks)
    printf(print_str)


@click.command('pnp')
@click.argument(
    'configfile',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
)
@click.option(
    '-c', '--check',
    is_flag=True,
    help="Only check the given config file but does not run it."
)
@click.option(
    '--no-log-probe',
    is_flag=True,
    help="Disables the automatic logging configuration probing."
)
@click.option(
    '--log',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Specify logging configuration to load."
)
@click.option(
    '--log-level',
    type=click.Choice(['DEBUG', 'INFO', 'WARNING', 'ERROR'], case_sensitive=False),
    default=None,
    help="Overrides the log level."
)
@click.version_option(version=__version__)
def main(configfile, check, log, log_level, no_log_probe):
    """Pull 'n' Push. Runs or checks the given CONFIGFILE"""
    printf(f"{fg.green}{bg.black}{PNP}{bg.rs}{fg.rs}")
    printf(f"{ef.bold}Welcome to {fg.green}pnp{fg.rs} @ {fg.green}{__version__}{rs.all}")
    printf()

    app = Application.from_file(configfile)
    config = app.config

    _print_api_config(config)
    _print_engine_config(config)
    _print_udf_config(config)
    _print_tasks_config(config)

    if not check:
        log_level_override = log_level or os.environ.get('LOG_LEVEL')
        probing_log_conf = [log]
        if not no_log_probe:
            probing_log_conf += [
                # logging.yaml in cwd
                DEFAULT_LOGGING_FILE_NAME,
                # logging.yaml in pnp config loc
                os.path.join(os.path.dirname(configfile), DEFAULT_LOGGING_FILE_NAME)
            ]
        logging_config_path = _setup_logging(
            *probing_log_conf,
            log_level_override=log_level_override
        )
        printf(f"{ef.bold}Logging{rs.all}\n{DSPACE}{fg.green}{logging_config_path}{fg.rs}")

        runner = Runner.choose_runner(app)
        runner.run()


if __name__ == '__main__':
    main()  # pylint: disable=no-value-for-parameter



"""Contains base stuff for user-defined functions in selector expressions."""

from abc import abstractmethod
from typing import Optional, Any

import cachetools  # type: ignore

from pnp.plugins import Plugin
from pnp.utils import (
    parse_duration_literal,
    DurationLiteral,
    make_hashable
)


class UserDefinedFunction(Plugin):
    """Base class for a user defined expression."""

    __REPR_FIELDS__ = ['throttle']

    MAX_CACHE_SIZE = 12

    def __init__(self, throttle: Optional[DurationLiteral] = None, **kwargs: Any):
        """
        Initializer.

        Args:
            throttle: If set to a valid duration literal (e.g. 5m) the return value of the
              called functions will be cached for the given amount of time.
        """
        super().__init__(**kwargs)
        self.throttle = throttle and parse_duration_literal(throttle)
        self._cache = None
        if self.throttle:
            self._cache = cachetools.TTLCache(self.MAX_CACHE_SIZE, ttl=self.throttle)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self._cache is None:
            return self.action(*args, **kwargs)

        hashable_args = (make_hashable(args), make_hashable(kwargs))
        try:
            return self._cache[hashable_args]
        except KeyError:
            res = self.action(*args, **kwargs)
            self._cache[hashable_args] = res
            return res

    @abstractmethod
    def action(self, *args: Any, **kwargs: Any) -> Any:
        """Actual definition of the hard-work of the user defined function."""
        raise NotImplementedError()



"""Contains a ping endpoint."""

from fastapi import FastAPI
from starlette.responses import PlainTextResponse

from .base import Endpoint


class Ping(Endpoint):
    """Ping endpoint. When requested just returns a simple pong."""

    async def endpoint(self) -> str:
        """Returns 'pong' when requested. Use this as a basic health endpoint.
        This will not do any additional checks (like database access, ...)
        but accessibility only."""
        return "pong"

    def attach(self, fastapi: FastAPI) -> None:
        """Attach the endpoint to the serving component."""
        fastapi.get(
            path="/ping",
            response_class=PlainTextResponse
        )(self.endpoint)



"""GPIO related plugins."""

from collections import Counter

from pnp import validator
from pnp.plugins.pull import SyncPull
from pnp.shared.gpio import (
    CONST_RISING, CONST_RISING_OPTIONS, CONST_FALLING_OPTIONS,
    CONST_SWITCH_OPTIONS, CONST_MOTION_OPTIONS, Callback, GPIOAdapter
)
from pnp.utils import (make_list)


class Watcher(SyncPull):
    """
    Listens for low/high state changes on the configured gpio pins.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/gpio.Watcher/index.md
    """
    __REPR_FIELDS__ = ['_mode_default', '_pins']

    EXTRA = 'gpio'

    def __init__(self, pins, default=CONST_RISING, **kwargs):
        super().__init__(**kwargs)
        self._mode_default = default
        validator.one_of(
            CONST_RISING_OPTIONS + CONST_FALLING_OPTIONS + CONST_SWITCH_OPTIONS
            + CONST_MOTION_OPTIONS,
            mode_default=self._mode_default
        )
        self._pins = [Callback.from_str(pin_str, default=default) for pin_str in make_list(pins)]
        _without_duplicate = set(self._pins)
        if len(_without_duplicate) != len(self._pins):
            diff = list((Counter(self._pins) - Counter(_without_duplicate)).elements())
            self.logger.warning(
                "You provided duplicate gpio pin configurations. Will ignore '%s'", diff
            )
            self._pins = _without_duplicate

    def _universal_callback(self, gpio_pin, event):
        self.logger.info("GPIO '%s' raised event '%s'", gpio_pin, event)
        self.notify(dict(gpio_pin=gpio_pin, event=event))

    def _pull(self):
        GPIO = GPIOAdapter()

        try:
            for pin in self._pins:
                pin.run(self._universal_callback)
            GPIO.apply()
            while not self.stopped:
                self._sleep()
        finally:
            for pin in self._pins:
                pin.stop()
            GPIO.cleanup()



"""Time database related push plugins."""

from pnp.plugins.push import SyncPush, enveloped


class InfluxPush(SyncPush):
    """
    Pushes the given `payload` to an influx database using the line `protocol`.
    You have to specify `host`, `port`, `user`, `password` and the `database`.

    The `protocol` is basically a string that will be augmented at push-time with data from the
    payload. E.g. {payload.metric},room={payload.location} value={payload.value} assumes that
    payload contains metric, location and value.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/timedb.InfluxPush/index.md
    """
    __REPR_FIELDS__ = ['database', 'host', 'port', 'protocol', 'user']

    def __init__(self, host, port, user, password, database, protocol, **kwargs):
        super().__init__(**kwargs)
        self.host = str(host)
        self.port = int(port)
        self.user = str(user)
        self.password = str(password)
        self.database = str(database)
        self.protocol = str(protocol)

    @enveloped
    def _push(self, envelope, payload):  # pylint: disable=arguments-differ
        points = [self.protocol.format(payload=payload)]
        self.logger.debug("Writing '%s' to influxdb", str(points))

        from influxdb import InfluxDBClient
        client = InfluxDBClient(self.host, self.port, self.user, self.password, self.database)
        client.write(points, {'db': self.database}, 204, 'line')

        return {'data': payload, **envelope} if envelope else payload



"""Contains api models that are globally useful."""
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from starlette.responses import JSONResponse
from starlette.status import HTTP_200_OK


class _EmptyResponse:
    """An empty json response."""
    def __call__(self, status_code: int = HTTP_200_OK) -> JSONResponse:
        return JSONResponse(status_code=status_code, content={})


EmptyResponse = _EmptyResponse()


class HealthResponse(BaseModel):
    """Basic health response."""

    success: bool = True



"""Contains a health endpoint."""

from fastapi import FastAPI

from pnp.api.models import HealthResponse
from .base import Endpoint


class Health(Endpoint):
    """When requested returns a health response with the general health of the application.
       Right now this does not perform any health checks, but it will in the future."""

    async def endpoint(self) -> HealthResponse:
        """When requested returns a health response with the general health of the application.
        Right now this does not perform any health checks, but it will in the future."""
        return HealthResponse()

    def attach(self, fastapi: FastAPI) -> None:
        """Attach the endpoint to the serving component."""
        fastapi.get(
            path="/health",
            response_model=HealthResponse
        )(self.endpoint)



"""Useful sensor stuff."""

import logging
from functools import partial

import pydantic
import schema as sc

from pnp import validator
from pnp.plugins import load_optional_module
from pnp.plugins.pull import PollingError, SyncPolling, SyncPull
from pnp.shared.sound import (
    WavFile,
    similarity_pearson,
    similarity_std,
    MODE_PEARSON,
    MODE_STD,
    ALLOWED_MODES
)
from pnp.utils import (
    parse_duration_literal,
    Cooldown
)


class DHT(SyncPolling):
    """
    Periodically polls a dht11 or dht22 (aka am2302) for temperature and humidity readings.
    Polling interval is controlled by `interval`.

    See Also:
        https://pnp.readthedocs.io/en/stable/plugins/index.html#sensor-dht
    """
    __REPR_FIELDS__ = ['data_gpio', 'device', 'humidity_offset', 'temp_offset']

    EXTRA = 'dht'

    def __init__(self, device='dht22', data_gpio=17, humidity_offset=0.0, temp_offset=0.0,
                 **kwargs):
        super().__init__(**kwargs)
        valid_devices = ['dht11', 'dht22', 'am2302']
        self.device = str(device).lower()
        validator.one_of(valid_devices, device=self.device)
        self.data_gpio = int(data_gpio)
        self.humidity_offset = float(humidity_offset)
        self.temp_offset = float(temp_offset)

    @staticmethod
    def _load_dht_package():
        try:
            import Adafruit_DHT as Pkg
        except ImportError:  # pragma: no cover
            logger = logging.getLogger(__name__)
            logger.warning("Adafruit_DHT package is not available - Using mock")
            from ...mocking import DHTMock as Pkg
        return Pkg

    def _poll(self):
        # Adafruit package is optional - import at the last moment
        pkg = self._load_dht_package()

        device_map = {'dht11': pkg.DHT11, 'dht22': pkg.DHT22, 'am2302': pkg.AM2302}
        humidity, temperature = pkg.read_retry(device_map[self.device], self.data_gpio)

        if humidity is None or temperature is None:
            raise PollingError("Failed to get '{}' @ GPIO {} readings...".format(
                self.device, str(self.data_gpio)))

        return {
            'humidity': round(float(humidity) + self.humidity_offset, 2),
            'temperature': round(float(temperature) + self.temp_offset, 2)
        }


class MiFlora(SyncPolling):
    """
    Periodically polls a xiaomi miflora plant sensor for sensor readings via btle.

    See Also:
        https://pnp.readthedocs.io/en/stable/plugins/index.html#sensor-miflora
    """
    __REPR_FIELDS__ = ['adapter', 'mac']

    EXTRA = 'miflora'

    def __init__(self, mac, adapter='hci0', **kwargs):
        super().__init__(**kwargs)
        self.mac = str(mac)
        self.adapter = str(adapter)
        self._poller = None

    def _find_backend(self):
        try:
            import bluepy.btle  # noqa: F401 pylint: disable=unused-import
            from btlewrap import BluepyBackend
            backend = BluepyBackend
        except ImportError:
            from btlewrap import GatttoolBackend
            backend = GatttoolBackend
        self.logger.debug("Miflora for %s is using %s backend", self.mac, backend.__name__)
        return backend

    def _init(self):
        if self._poller:
            # We already got a miflora poller -> abort
            return

        mfp = load_optional_module('miflora.miflora_poller', self.EXTRA)
        self._poller = mfp.MiFloraPoller(
            mac=self.mac, adapter=self.adapter, backend=self._find_backend()
        )
        self.logger.debug("Initialization for %s finished", self.mac)

    def _connect_sensor(self):
        assert self._poller

        from btlewrap import BluetoothBackendException
        try:
            self.logger.debug("Reading miflora sensor: %s", self.mac)
            self._poller.fill_cache()
        except IOError as ioerr:
            raise PollingError() from ioerr
        except BluetoothBackendException as bterror:
            raise PollingError() from bterror

    def _get_sensor_readings(self):
        assert self._poller

        mfp = load_optional_module('miflora.miflora_poller', self.EXTRA)
        reading_params = [mfp.MI_CONDUCTIVITY, mfp.MI_LIGHT, mfp.MI_MOISTURE,
                          mfp.MI_TEMPERATURE, mfp.MI_BATTERY]
        res = {para: self._poller.parameter_value(para) for para in reading_params}
        return {**res, **{'firmware': self._poller.firmware_version()}}

    def _poll(self):
        self._init()
        self._connect_sensor()
        readings = self._get_sensor_readings()
        self.logger.info("Miflora readings for '%s': %s", self.mac, readings)
        return readings


# pylint: disable=too-many-instance-attributes
class Sound(SyncPull):
    """
    Listens to the microphone in realtime and searches the stream for a specific sound pattern.
    Practical example: I use this plugin to recognize my doorbell without tampering with
    the electrical device ;-)

    See Also:
        https://pnp.readthedocs.io/en/stable/plugins/index.html#sensor-sound
    """
    class WavConfig(pydantic.BaseModel):
        """Wav configuration entity."""
        wav_file: WavFile
        mode: str
        offset: float
        cooldown_period: int
        cooldown_event: bool
        notify: Cooldown

        class Config:
            """Pydantic config."""
            arbitrary_types_allowed = True

        @pydantic.validator('mode')
        @classmethod
        def _check_mode(cls, mode: str) -> str:
            validator.one_of(ALLOWED_MODES, mode=mode)
            return mode

    class WavConfigSchema:
        """Wav configuration configuration schema."""

        CONF_COOLDOWN = 'cooldown'
        CONF_COOLDOWN_PERIOD = 'period'
        CONF_COOLDOWN_EVENT = 'emit_event'
        CONF_PATH = 'path'
        CONF_MODE = 'mode'
        CONF_OFFSET = 'offset'

        DEFAULT_COOLDOWN_PERIOD = 10
        DEFAULT_COOLDOWN = {
            CONF_COOLDOWN_PERIOD: DEFAULT_COOLDOWN_PERIOD,
            CONF_COOLDOWN_EVENT: False
        }
        DEFAULT_MODE = MODE_PEARSON
        DEFAULT_OFFSET = 0.0

        WAV_FILE_SCHEMA = sc.Schema({
            CONF_PATH: sc.Use(str),
            sc.Optional(CONF_MODE, default=DEFAULT_MODE): sc.Use(str),
            sc.Optional(CONF_OFFSET, default=DEFAULT_OFFSET): sc.Use(float),
            sc.Optional(CONF_COOLDOWN, default=DEFAULT_COOLDOWN): {
                sc.Optional(CONF_COOLDOWN_PERIOD, default=DEFAULT_COOLDOWN_PERIOD):
                    sc.Use(parse_duration_literal),
                sc.Optional(CONF_COOLDOWN_EVENT, default=False): sc.Use(bool)
            }
        })

        WAV_FILE_LIST_SCHEMA = sc.Schema([WAV_FILE_SCHEMA])

        @classmethod
        def from_dict(cls, dct_config, base_path, notify_fun, cooldown_fun):
            """Loads a wav configuration from a dictionary."""
            config = cls.WAV_FILE_SCHEMA.validate(dct_config)
            wav_file = WavFile.from_path(config[cls.CONF_PATH], base_path)
            cooldown_cfg = config[cls.CONF_COOLDOWN]
            cooldown_cb = None
            if cooldown_cfg[cls.CONF_COOLDOWN_EVENT]:
                cooldown_cb = partial(cooldown_fun, file_name=wav_file.file_name)
            return Sound.WavConfig(
                wav_file=wav_file,
                mode=config[cls.CONF_MODE],
                offset=config[cls.CONF_OFFSET],
                cooldown_period=cooldown_cfg[cls.CONF_COOLDOWN_PERIOD],
                cooldown_event=cooldown_cfg[cls.CONF_COOLDOWN_EVENT],
                notify=Cooldown(
                    notify_fun,
                    cool_down=cooldown_cfg[cls.CONF_COOLDOWN_PERIOD],
                    cool_down_callback=cooldown_cb
                )
            )

        @classmethod
        def from_list(cls, list_config, base_path, notify_fun, cooldown_fun):
            """Loads multiple wav configurations from a list."""
            validated = cls.WAV_FILE_LIST_SCHEMA.validate(list_config)
            return [
                cls.from_dict(wav_file, base_path, notify_fun, cooldown_fun)
                for wav_file in validated
            ]

    __REPR_FIELDS__ = ['device_index', 'ignore_overflow', 'wav_files']

    EXTRA = 'sound'

    RATE = 44100
    CHUNK_SIZE = 1024 * 4
    PEARSON_THRESHOLD = 0.5
    STD_THRESHOLD = 1.4

    def __init__(self, wav_files, device_index=None, ignore_overflow=False,
                 **kwargs):
        super().__init__(**kwargs)
        self.wav_files = self.WavConfigSchema.from_list(
            wav_files, self.base_path, self.notify, self._on_cooldown
        )
        self.device_index = device_index and int(device_index)
        self.ignore_overflow = bool(ignore_overflow)

    def _check_dependencies(self):
        load_optional_module('numpy', self.EXTRA)
        load_optional_module('pyaudio', self.EXTRA)
        load_optional_module('scipy', self.EXTRA)
        load_optional_module('scipy.io.wavfile', self.EXTRA)
        load_optional_module('scipy.stats.stats', self.EXTRA)

    def _similarity(self, buffer, config: WavConfig):
        if config.mode == MODE_PEARSON:
            return similarity_pearson(
                buffer,
                config.wav_file,
                self.RATE,
                self.PEARSON_THRESHOLD + config.offset
            )
        if config.mode == MODE_STD:
            return similarity_std(
                buffer,
                config.wav_file,
                self.RATE,
                self.STD_THRESHOLD + config.offset
            )
        raise ValueError("The given mode '{}' is unsupported".format(config.mode))

    def _on_cooldown(self, file_name):
        self.notify({
            'type': 'cooldown',
            'sound': file_name
        })

    def _pull(self):
        self._check_dependencies()
        np = load_optional_module('numpy', self.EXTRA)
        pyaudio = load_optional_module('pyaudio', self.EXTRA)

        pa = pyaudio.PyAudio()
        stream = pa.open(
            format=pyaudio.paInt16,
            input_device_index=self.device_index,
            channels=1,
            rate=self.RATE,
            input=True,
            frames_per_buffer=self.CHUNK_SIZE
        )
        try:
            buffers = [None for _ in self.wav_files]  # Init buffer foreach wav_file
            while not self.stopped:
                data = np.fromstring(stream.read(
                    self.CHUNK_SIZE,
                    exception_on_overflow=not self.ignore_overflow
                ), dtype=np.int16)

                for i, buffer in enumerate(buffers):
                    config = self.wav_files[i]
                    buffer = self._process_single(buffer, data, np, config)
                    buffers[i] = buffer
        finally:
            stream.close()
            pa.terminate()

            # Emit any pending cooldown events
            for config in self.wav_files:
                config.notify.execute_now()

    def _process_single(self, buffer, data, np, config):
        wav_file = config.wav_file
        N = wav_file.signal_length
        buffer = data if buffer is None else np.concatenate((buffer, data), axis=None)
        lbuf = len(buffer)
        if lbuf >= N:
            self.logger.debug("Buffer (%s): %s (Buffer) >= %s (Wav)",
                              wav_file.file_name, lbuf, N)
            flag, corrcoef, threshold = self._similarity(buffer, config)
            self.logger.debug("Correlation (%s): %s >= %s = %s",
                              wav_file.file_name, corrcoef, threshold, flag)
            if flag:
                config.notify({
                    'type': 'sound',
                    'sound': wav_file.file_name,
                    'corrcoef': corrcoef,
                    'threshold': threshold
                })
                buffer = None
            else:
                buffer = buffer[int(len(buffer) * 0.75):]

        return buffer



"""Contains runner to run an application."""

import asyncio
import os
import signal

import uvicorn

from pnp.api import APINotConfiguredError
from pnp.app import Application
from pnp.utils import Loggable


class Runner(Loggable):
    """Runner to run a basic non-api application."""
    def __init__(self, app: Application):
        self.app = app

    async def _main_loop(self) -> None:
        while self.app.engine.is_running:
            await asyncio.sleep(0.1)

    def run(self) -> None:
        """Run the application."""
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.app.engine.start(self.app.tasks))
        try:
            loop.run_until_complete(self._main_loop())
        except KeyboardInterrupt:
            try:
                self.logger.info(
                    "Attempting graceful shutdown, press Ctrl+C again to exit forcefully..."
                )
                loop.run_until_complete(self.app.engine.stop())
            except KeyboardInterrupt:
                self.logger.info("Forceful exit")

    @classmethod
    def choose_runner(cls, app: Application) -> 'Runner':
        """Factory method to choose the correct runner for the given application."""
        if app.api is not None:
            return APIAwareRunner(app)
        return cls(app)


class APIAwareRunner(Runner):
    """Runner that knows how to handle a fastapi api."""
    async def _on_engine_stopped(self):
        # The only way I know to programmatically stop uvicorn politely
        # without messing with uvicorn internals
        os.kill(os.getpid(), signal.SIGTERM)

    def run(self):
        api = self.app.api
        if not api:
            raise APINotConfiguredError()
        engine = self.app.engine
        engine.on_stopped_callback = self._on_engine_stopped

        @api.fastapi.on_event("startup")  # type: ignore
        async def _on_startup() -> None:
            await engine.start(self.app.tasks)

        @api.fastapi.on_event("shutdown")  # type: ignore
        async def _on_shutdown() -> None:
            await engine.stop()

        uvicorn.run(
            api.fastapi,
            host="0.0.0.0",
            port=self.app.config.api.port,
            log_config=None,
            log_level="debug"
        )



"""Basic push plugins."""

from functools import partial
from typing import Union, Any

from pnp import validator
from pnp.plugins.push import enveloped, SyncPush, AsyncPush
from pnp.shared.exc import TemplateError
from pnp.utils import parse_duration_literal, make_list, parse_duration_literal_float


class Echo(AsyncPush):
    """
    This push simply logs the `payload` via the `logging` module.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/simple.Echo/index.md

    Examples:

        >>> import asyncio
        >>> loop = asyncio.get_event_loop()
        >>> dut = Echo(name="echo_push")
        >>> loop.run_until_complete(dut.push("I will be logged"))
        'I will be logged'
    """

    @enveloped
    async def _push(self, envelope, payload):  # pylint: disable=arguments-differ
        self.logger.info("Got '%s' with envelope '%s'", payload, envelope)
        # Payload as is. With envelope (if any)
        return {'data': payload, **envelope} if envelope else payload


class Nop(AsyncPush):
    """
    Executes no operation at all. A call to push(...) just returns the payload.
    This push is useful when you only need the power of the selector for dependent pushes.

    Nop = No operation OR No push ;-)

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/simple.Nop/index.md

    Examples:

        >>> import asyncio
        >>> loop = asyncio.get_event_loop()
        >>> dut = Nop(name="nop_push")
        >>> loop.run_until_complete(dut.push('I will be returned unaltered'))
        'I will be returned unaltered'
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.last_payload = None

    async def _push(self, payload):
        self.last_payload = payload
        return payload


class Wait(AsyncPush):
    """
    Performs a sleep operation and wait for some time to go by.
    """
    __REPR_FIELDS__ = ['waiting_interval']

    def __init__(self, wait_for: Union[str, float, int], **kwargs: Any):
        super().__init__(**kwargs)

        self.waiting_interval = parse_duration_literal_float(wait_for)

    async def _push(self, payload):
        import asyncio
        await asyncio.sleep(self.waiting_interval)
        return payload


class Execute(SyncPush):
    """
    Executes a command with given arguments in a shell of the operating system.
    Both `command` and `args` may include placeholders (e.g. `{{placeholder}}`) which are injected
    at runtime by passing the specified payload after selector transformation.

    Will return the exit code of the command and optionally the output from stdout and stderr.
    """
    __REPR_FIELDS__ = ['_args', '_capture', '_command', '_cwd', '_timeout']

    def __init__(self, command, args=None, cwd=None, capture=True, timeout="5s", **kwargs):
        super().__init__(**kwargs)
        self._command = str(command) if command else None
        self._args = self._parse_args(args)
        self._cwd = str(cwd) if cwd else self.base_path
        if self._cwd:
            validator.is_directory(cwd=self._cwd)
        self._capture = bool(capture)
        self._timeout = timeout and parse_duration_literal(timeout)

    @staticmethod
    def _render_jinja_template(template, subs):
        from jinja2 import StrictUndefined
        from jinja2 import Template
        from jinja2.exceptions import UndefinedError
        try:
            tpl = Template(template, undefined=StrictUndefined)
            return tpl.render(**(subs or {}))
        except UndefinedError as exc:
            raise TemplateError('Error when rendering template in TemplatedExecute') from exc

    @staticmethod
    def _parse_args(val):
        args = make_list(val)
        if not args:
            return None
        return [str(arg) for arg in args]

    def _serialize_args(self, transform_fun=None, add_quotes=True):
        def escape_fun(arg):
            return str(arg).replace('"', '\\"')

        def quotes_fun(arg):
            if arg is None or not str(arg) or str(arg).strip() == 'None':
                return ''
            return '"{}"'.format(escape_fun(arg)) if add_quotes else str(arg)

        args = self._args
        if not args:
            return None
        if transform_fun:
            args = [transform_fun(arg) for arg in args]
        # Do argument quoting
        args = [quotes_fun(arg) for arg in args]
        # Remove empty args
        args = [arg for arg in args if arg != '']
        return " ".join(args)

    def _execute(self, command_str):
        def _output(response):
            return [line.strip('\n\r') for line in response]

        import subprocess
        self.logger.info("Running command in shell: %s", command_str)
        proc = subprocess.Popen(
            args=command_str,
            shell=True,
            cwd=self._cwd,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        try:
            if self._timeout:
                proc.wait(timeout=int(self._timeout))

            res = dict(return_code=proc.returncode)
            if self._capture:
                res['stdout'] = _output(proc.stdout)
                res['stderr'] = _output(proc.stderr)
            return res
        finally:
            proc.stdout.close()
            proc.stderr.close()

    def _push(self, payload):
        if isinstance(payload, dict):
            subs = payload
        else:
            subs = dict(data=payload, payload=payload)

        command_str = self._render_jinja_template(self._command, subs=subs)
        if self._args:
            args = self._serialize_args(
                transform_fun=partial(self._render_jinja_template, subs=subs),
                add_quotes=True
            )
            command_str = "{command_str} {args}".format(command_str=command_str, args=args)

        return self._execute(command_str)



"""Storage related push plugins."""
from typing import Optional, Any

from typeguard import typechecked

from pnp.plugins import load_optional_module
from pnp.plugins.push import SyncPush, enveloped, parse_envelope, drop_envelope
from pnp.utils import get_bytes


class Dropbox(SyncPush):
    """
    Uploads provided file to the specified dropbox account.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/storage.Dropbox/index.md
    """
    __REPR_FIELDS__ = ['create_shared_link', 'target_file_name']

    EXTRA = 'dropbox'

    def __init__(
        self, api_key: str, target_file_name: Optional[str] = None,
        create_shared_link: bool = True, **kwargs: Any
    ):
        super().__init__(**kwargs)
        self.api_key = str(api_key)
        self.target_file_name = target_file_name and str(target_file_name)
        self.create_shared_link = bool(create_shared_link)

    @staticmethod
    @typechecked
    def _sanitze_target_file_name(val: str) -> str:
        if not val.startswith('/'):
            return '/' + val
        return val

    @staticmethod
    def _make_raw_file_url(shared_url):
        try:
            import urlparse
            from urllib import urlencode
        except ImportError:  # For Python 3
            import urllib.parse as urlparse
            from urllib.parse import urlencode

        url_parts = list(urlparse.urlparse(shared_url))
        # Replace all query params with ?raw=1
        url_parts[4] = urlencode({'raw': '1'})
        return urlparse.urlunparse(url_parts)

    @enveloped
    @parse_envelope('target_file_name')
    @drop_envelope
    def _push(self, target_file_name, payload):  # pylint: disable=arguments-differ
        target_file_name = self._sanitze_target_file_name(target_file_name)

        # Upload file or stream to dropbox
        dropbox = load_optional_module('dropbox', self.EXTRA)
        dbx = dropbox.Dropbox(self.api_key, timeout=None)
        fcontent = get_bytes(payload)  # Might be a file or a stream
        self.logger.info("Uploading file to dropbox '%s'", target_file_name)
        metadata = dbx.files_upload(
            fcontent,
            target_file_name,
            mode=dropbox.files.WriteMode("overwrite")
        )

        res = dict(
            name=metadata.name,
            id=metadata.id,
            content_hash=metadata.content_hash,
            size=metadata.size,
            path=metadata.path_display,
            shared_link=None,
            raw_link=None
        )
        if self.create_shared_link:
            # create a shared link
            shared_link = dbx.sharing_create_shared_link(target_file_name)
            res['shared_link'] = shared_link.url
            res['raw_link'] = self._make_raw_file_url(shared_link.url)

        return res
